from typing import Any, Callable, ClassVar

import numpy as np
import torch

import lucid
from lucid.test.core import TensorOpWithTorchBase, TensorOpTorchCase
from lucid.test._utils.test_utils import _UtilsTupleTorchOpBase


def _unwrap_callable(operation: Any) -> Callable[..., Any]:
    if isinstance(operation, staticmethod):
        return operation.__func__
    if callable(operation):
        return operation
    raise TypeError(f"Expected callable operation, got: {type(operation)!r}")


class _LinalgUnaryTorchBase(TensorOpWithTorchBase):
    _collect_nested_test_classes = True
    case_name: ClassVar[str]
    input: ClassVar[Any]

    forward_op: ClassVar[Callable[..., Any]]
    torch_forward: ClassVar[Callable[..., Any] | None]
    expected_forward: ClassVar[Any]

    requires_grad: ClassVar[bool] = True
    dtype: ClassVar[Any | None] = None

    compare_grad_with_torch: ClassVar[bool] = True
    check_finite: ClassVar[bool] = True
    expected_input_grads: ClassVar[tuple[object | None, ...] | None] = None

    rtol: ClassVar[float] = 1e-7
    atol: ClassVar[float] = 1e-8

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu",)

    def _build_case(self) -> TensorOpTorchCase:
        forward_op = vars(type(self)).get("forward_op")
        torch_forward = vars(type(self)).get("torch_forward")

        if isinstance(forward_op, staticmethod):
            forward_op = forward_op.__func__
        if isinstance(torch_forward, staticmethod):
            torch_forward = torch_forward.__func__

        return TensorOpTorchCase(
            name=self.case_name,
            build_inputs=lambda _case: (
                self.tensor(
                    self.input, requires_grad=self.requires_grad, dtype=self.dtype
                ),
            ),
            forward_op=forward_op,
            expected_forward=self.expected_forward,
            expected_input_grads=self.expected_input_grads,
            requires_grad=(self.requires_grad,),
            check_finite=self.check_finite,
            torch_forward=cast_callable(torch_forward),
            compare_grad_with_torch=self.compare_grad_with_torch,
            rtol=self.rtol,
            atol=self.atol,
        )

    def tensor_op_cases(self) -> tuple[TensorOpTorchCase, ...]:
        return (self._build_case(),)


class _LinalgBinaryTorchBase(TensorOpWithTorchBase):
    _collect_nested_test_classes = True
    case_name: ClassVar[str]
    left: ClassVar[Any]
    right: ClassVar[Any]

    forward_op: ClassVar[Callable[..., Any]]
    torch_forward: ClassVar[Callable[..., Any] | None]
    expected_forward: ClassVar[Any]

    requires_grad: ClassVar[bool | tuple[bool, bool]] = True
    compare_grad_with_torch: ClassVar[bool] = True

    check_finite: ClassVar[bool] = True
    expected_input_grads: ClassVar[tuple[object | None, ...] | None] = None
    dtype: ClassVar[Any | None] = None

    rtol: ClassVar[float] = 1e-7
    atol: ClassVar[float] = 1e-8

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu",)

    def _build_case(self) -> TensorOpTorchCase:
        forward_op = vars(type(self)).get("forward_op")
        torch_forward = vars(type(self)).get("torch_forward")

        if isinstance(forward_op, staticmethod):
            forward_op = forward_op.__func__
        if isinstance(torch_forward, staticmethod):
            torch_forward = torch_forward.__func__

        requires_grad = self.requires_grad
        if isinstance(requires_grad, bool):
            requires_grad = (requires_grad, requires_grad)

        return TensorOpTorchCase(
            name=self.case_name,
            build_inputs=lambda _case: (
                self.tensor(
                    self.left, requires_grad=requires_grad[0], dtype=self.dtype
                ),
                self.tensor(
                    self.right, requires_grad=requires_grad[1], dtype=self.dtype
                ),
            ),
            forward_op=forward_op,
            expected_forward=self.expected_forward,
            expected_input_grads=self.expected_input_grads,
            requires_grad=requires_grad,
            check_finite=self.check_finite,
            torch_forward=cast_callable(torch_forward),
            compare_grad_with_torch=self.compare_grad_with_torch,
            rtol=self.rtol,
            atol=self.atol,
        )

    def tensor_op_cases(self) -> tuple[TensorOpTorchCase, ...]:
        return (self._build_case(),)


def cast_callable(value: Callable[..., Any] | None) -> Callable[..., Any] | None:
    if value is None:
        return None
    return _unwrap_callable(value)


class TestInv(_LinalgUnaryTorchBase):
    case_name = "inv"
    input = np.array([[4.0, 7.0], [2.0, 6.0]], dtype=np.float64)
    forward_op = staticmethod(lambda a: lucid.linalg.inv(a))
    torch_forward = staticmethod(torch.linalg.inv)
    expected_forward = np.linalg.inv(input)
    compare_grad_with_torch = False


class TestDet(_LinalgUnaryTorchBase):
    case_name = "det"
    input = np.array([[3.0, 2.0], [1.0, 4.0]], dtype=np.float64)
    forward_op = staticmethod(lambda a: lucid.linalg.det(a))
    torch_forward = staticmethod(torch.linalg.det)
    expected_forward = np.linalg.det(input)
    compare_grad_with_torch = False


class TestSolve(_LinalgBinaryTorchBase):
    case_name = "solve"
    left = np.array([[3.0, 1.0], [1.0, 2.0]], dtype=np.float64)
    right = np.array([[9.0], [8.0]], dtype=np.float64)
    forward_op = staticmethod(lambda a, b: lucid.linalg.solve(a, b))
    torch_forward = staticmethod(torch.linalg.solve)
    expected_forward = np.linalg.solve(left, right)
    compare_grad_with_torch = False


class TestCholesky(_LinalgUnaryTorchBase):
    case_name = "cholesky"
    input = np.array([[4.0, 1.0], [1.0, 3.0]], dtype=np.float64)
    forward_op = staticmethod(lambda a: lucid.linalg.cholesky(a))
    torch_forward = staticmethod(torch.linalg.cholesky)
    expected_forward = np.linalg.cholesky(input)
    compare_grad_with_torch = False


class TestNorm(_LinalgUnaryTorchBase):
    case_name = "norm"
    input = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype=np.float64)
    forward_op = staticmethod(
        lambda a: lucid.linalg.norm(a, ord=1, axis=1, keepdims=True)
    )
    torch_forward = staticmethod(
        lambda a: torch.linalg.norm(a, ord=1, dim=1, keepdim=True)
    )
    expected_forward = np.array([6.0, 15.0], dtype=np.float64).reshape(2, 1)


class TestMatrixPower(_LinalgUnaryTorchBase):
    case_name = "matrix_power"
    input = np.array([[2.0, 1.0], [0.0, 3.0]], dtype=np.float64)
    forward_op = staticmethod(lambda a: lucid.linalg.matrix_power(a, n=3))
    torch_forward = staticmethod(lambda a: torch.linalg.matrix_power(a, n=3))
    expected_forward = np.linalg.matrix_power(input, 3)
    compare_grad_with_torch = False


class TestPInv(_LinalgUnaryTorchBase):
    case_name = "pinv"
    input = np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64)
    forward_op = staticmethod(lambda a: lucid.linalg.pinv(a))
    torch_forward = staticmethod(torch.linalg.pinv)
    expected_forward = np.linalg.pinv(input)


class _LinalgTupleTorchBase(_UtilsTupleTorchOpBase):
    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu",)


class TestQR(_LinalgTupleTorchBase):
    case_name = "qr"
    input = np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64)
    inputs = (input,)
    forward_op = staticmethod(lambda a: lucid.linalg.qr(a))
    expected_forward = np.linalg.qr(input)
    compare_grad_with_torch = False
    check_finite = True


class TestSVD(_LinalgTupleTorchBase):
    case_name = "svd"
    input = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype=np.float64)
    inputs = (input,)
    forward_op = staticmethod(lambda a: lucid.linalg.svd(a))
    expected_forward = np.linalg.svd(input)
    compare_grad_with_torch = False
    check_finite = True


class TestEig(_LinalgTupleTorchBase):
    case_name = "eig"
    input = np.array([[2.0, 0.0], [0.0, 1.0]], dtype=np.float64)
    inputs = (input,)
    forward_op = staticmethod(lambda a: lucid.linalg.eig(a))
    expected_forward = np.linalg.eig(input)
    compare_grad_with_torch = False
    check_finite = True



import importlib
import pytest

try:
    importlib.import_module("lucid")
    _LUCID_IMPORT_ERROR = None
except Exception as exc:
    _LUCID_IMPORT_ERROR = exc


@pytest.fixture(scope="session", autouse=True)
def require_lucid_package() -> None:
    if _LUCID_IMPORT_ERROR is not None:
        pytest.skip(
            f"lucid import failed during collection: {_LUCID_IMPORT_ERROR}",
            allow_module_level=True,
        )


def pytest_collection_modifyitems(
    config: pytest.Config, items: list[pytest.Item]
) -> None:
    _ = config
    training_test_file = "test_training.py"
    prior_items: list[pytest.Item] = []
    trailing_items: list[pytest.Item] = []

    for item in items:
        file_path, _, _ = item.location
        if file_path.endswith(training_test_file):
            trailing_items.append(item)
        else:
            prior_items.append(item)

    if trailing_items:
        items[:] = prior_items + trailing_items



from collections import defaultdict
from typing import Any, Iterable
from abc import ABC, abstractmethod
import copy

import lucid.nn as nn

from lucid.types import _OptimClosure


class Optimizer(ABC):
    def __init__(
        self, params: Iterable[nn.Parameter], defaults: dict[str, Any]
    ) -> None:
        if not isinstance(params, Iterable):
            raise TypeError("params should be an iterable of Parameters.")

        param_list = list(params)
        for p in param_list:
            if not isinstance(p, nn.Parameter):
                raise TypeError(f"Expected nn.Parameter, got {type(p).__name__}.")

        self.defaults: dict[str, Any] = dict(defaults)
        self.param_groups: list[dict[str, Any]] = self.param_groups_setup(
            param_list, self.defaults
        )
        self.state: dict[nn.Parameter, dict[str, Any]] = defaultdict(dict)

    @abstractmethod
    def step(self, closure: _OptimClosure | None = None) -> Any | None:
        raise NotImplementedError("The step method must be implemented by subclasses.")

    def zero_grad(self) -> None:
        for group in self.param_groups:
            for param in group["params"]:
                if isinstance(param, nn.Parameter):
                    param.zero_grad()

    def param_groups_setup(
        self, params: list[nn.Parameter], defaults: dict[str, Any]
    ) -> list[dict[str, Any]]:
        return [{"params": list(params), **defaults}]

    def add_param_group(self, param_group: dict[str, Any]) -> None:
        if "params" not in param_group:
            raise ValueError("param_group must have a 'params' key.")

        params = list(param_group["params"])
        if len(params) == 0:
            raise ValueError("param_group['params'] must be non-empty.")

        for p in params:
            if not isinstance(p, nn.Parameter):
                raise TypeError(
                    f"Expected nn.Parameter in param_group, got {type(p).__name__}."
                )

        existing = set()
        for g in self.param_groups:
            existing.update(g["params"])

        if any(p in existing for p in params):
            raise ValueError("Some parameters appear in more than one parameter group.")

        filled = {
            **self.defaults,
            **{k: v for k, v in param_group.items() if k != "params"},
        }
        filled["params"] = params
        self.param_groups.append(filled)

    def _flat_params(self) -> list[nn.Parameter]:
        flat: list[nn.Parameter] = []
        for g in self.param_groups:
            flat.extend(g["params"])

        return flat

    def state_dict(self) -> dict:
        param_to_idx: dict[nn.Parameter, int] = {}
        for idx, p in enumerate(self._flat_params()):
            if p not in param_to_idx:
                param_to_idx[p] = idx

        packed_state: dict[int, dict[str, Any]] = {}
        for p, st in self.state.items():
            if p in param_to_idx:
                packed_state[param_to_idx[p]] = copy.deepcopy(st)

        packed_groups: list[dict[str, Any]] = []
        for g in self.param_groups:
            new_g: dict[str, Any] = {}

            for k, v in g.items():
                if k == "params":
                    new_g[k] = [param_to_idx[p] for p in v]
                else:
                    new_g[k] = copy.deepcopy(v)

            packed_groups.append(new_g)

        return {"state": packed_state, "param_groups": packed_groups}

    def load_state_dict(self, state_dict: dict) -> None:
        if (
            not isinstance(state_dict, dict)
            or "state" not in state_dict
            or "param_groups" not in state_dict
        ):
            raise TypeError("Invalid state_dict format for Optimizer.")

        saved_groups = state_dict["param_groups"]
        saved_state = state_dict["state"]

        current_params = self._flat_params()
        n_current = len(current_params)

        new_groups: list[dict[str, Any]] = []
        for sg in saved_groups:
            if "params" not in sg:
                raise KeyError("Saved param_group missing 'params'.")
            indices: list[int] = list(sg["params"])

            if any(i < 0 or i >= n_current for i in indices):
                raise IndexError("Saved state refers to parameter index out of range.")

            params = [current_params[i] for i in indices]
            ng = {
                k: (params if k == "params" else copy.deepcopy(v))
                for k, v in sg.items()
            }
            new_groups.append(ng)

        self.param_groups = new_groups

        self.state = defaultdict(dict)
        for i, p in enumerate(self._flat_params()):
            if i in saved_state:
                self.state[p] = copy.deepcopy(saved_state[i])

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.defaults})"



from typing import (
    Any,
    Callable,
    ItemsView,
    Iterator,
    KeysView,
    Self,
    Type,
    TypeVar,
    ValuesView,
    overload,
)
from collections import OrderedDict

try:
    from tqdm.auto import tqdm
except Exception:
    tqdm = None

from lucid._tensor import Tensor
from lucid.types import (
    _ArrayOrScalar,
    _BackwardHook,
    _DeviceType,
    _ForwardHook,
    _ForwardHookKwargs,
    _ForwardPreHook,
    _ForwardPreHookKwargs,
    _FullBackwardHook,
    _FullBackwardPreHook,
    _LoadStateDictPostHook,
    _LoadStateDictPreHook,
    _NumPyArray,
    _StateDictHook,
    _StateDictPreHook,
)

import lucid.nn as nn

from lucid._jit.api import JITModule

__all__ = [
    "Module",
    "Sequential",
    "ModuleList",
    "ModuleDict",
    "ParameterList",
    "ParameterDict",
    "auto_repr",
    "set_state_dict_pass_attr",
]


class Module:
    _registry_map: dict[Type, OrderedDict[str, Any]] = {}
    _alt_name: str = ""

    def __init__(self) -> None:
        self._parameters: OrderedDict[str, nn.Parameter]
        self._buffers: OrderedDict[str, nn.Buffer]
        self._modules: OrderedDict[str, Self]

        object.__setattr__(self, "_parameters", OrderedDict())
        object.__setattr__(self, "_buffers", OrderedDict())
        object.__setattr__(self, "_modules", OrderedDict())

        self.training = True
        self.device: _DeviceType = "cpu"

        self._forward_pre_hooks: list[
            tuple[_ForwardPreHook | _ForwardPreHookKwargs, bool]
        ] = []
        self._forward_hooks: list[tuple[_ForwardHook | _ForwardHookKwargs, bool]] = []

        self._backward_hooks: list[_BackwardHook] = []
        self._full_backward_pre_hooks: list[_FullBackwardPreHook] = []
        self._full_backward_hooks: list[_FullBackwardHook] = []

        self._state_dict_pre_hooks: list[_StateDictPreHook] = []
        self._state_dict_hooks: list[_StateDictHook] = []

        self._load_state_dict_pre_hooks: list[_LoadStateDictPreHook] = []
        self._load_state_dict_post_hooks: list[_LoadStateDictPostHook] = []

        self._state_dict_pass_attr = set()

    def __setattr__(self, name: str, value: Any) -> None:
        registry_map: dict[Type, OrderedDict[str, Any]] = {
            nn.Parameter: self._parameters,
            nn.Buffer: self._buffers,
            Module: self._modules,
        }

        target_registry = None
        for cls, registry in registry_map.items():
            if isinstance(value, cls):
                target_registry = registry
                break

        if target_registry is not None:
            for registry in registry_map.values():
                if registry is not target_registry and name in registry:
                    del registry[name]
            target_registry[name] = value
        else:
            for registry in registry_map.values():
                if name in registry:
                    del registry[name]

        super().__setattr__(name, value)

    def setattr_raw(self, name: str, value: Any) -> None:
        object.__setattr__(self, name, value)

    def add_module(self, name: str, module: Self) -> None:
        if not isinstance(module, Module) and module is not None:
            raise TypeError(f"{module} is not a Module.")

        self.__setattr__(name, module)

    def register_parameter(self, name: str, param: nn.Parameter | None) -> None:
        if not isinstance(param, nn.Parameter) and param is not None:
            raise TypeError(f"{param} is not a nn.Parameter.")

        self.__setattr__(name, param)

    def register_buffer(
        self,
        name: str,
        buffer: nn.Buffer | _ArrayOrScalar | None,
        dtype: type | None = None,
    ) -> None:
        if buffer is not None:
            if not isinstance(buffer, nn.Buffer):
                buffer = nn.Buffer(buffer, dtype=dtype, device=self.device)

        self.__setattr__(name, buffer)

    def register_forward_pre_hook(
        self,
        hook: _ForwardPreHook | _ForwardPreHookKwargs,
        *,
        with_kwargs: bool = False,
    ) -> Callable:
        self._forward_pre_hooks.append((hook, with_kwargs))
        return lambda: self._forward_pre_hooks.remove((hook, with_kwargs))

    def register_forward_hook(
        self, hook: _ForwardHook | _ForwardHookKwargs, *, with_kwargs: bool = False
    ) -> Callable:
        self._forward_hooks.append((hook, with_kwargs))
        return lambda: self._forward_hooks.remove((hook, with_kwargs))

    def register_backward_hook(self, hook: _BackwardHook) -> Callable:
        self._backward_hooks.append(hook)
        return lambda: self._backward_hooks.remove(hook)

    def register_full_backward_pre_hook(self, hook: _FullBackwardPreHook) -> Callable:
        self._full_backward_pre_hooks.append(hook)
        return lambda: self._full_backward_pre_hooks.remove(hook)

    def register_full_backward_hook(self, hook: _FullBackwardHook) -> Callable:
        self._full_backward_hooks.append(hook)
        return lambda: self._full_backward_hooks.remove(hook)

    def register_state_dict_pre_hook(self, hook: _StateDictPreHook) -> Callable:
        self._state_dict_pre_hooks.append(hook)
        return lambda: self._state_dict_pre_hooks.remove(hook)

    def register_state_dict_hook(self, hook: _StateDictHook) -> Callable:
        self._state_dict_hooks.append(hook)
        return lambda: self._state_dict_hooks.remove(hook)

    def register_load_state_dict_pre_hook(
        self, hook: _LoadStateDictPreHook
    ) -> Callable:
        self._load_state_dict_pre_hooks.append(hook)
        return lambda: self._load_state_dict_pre_hooks.remove(hook)

    def register_load_state_dict_post_hook(
        self, hook: _LoadStateDictPostHook
    ) -> Callable:
        self._load_state_dict_post_hooks.append(hook)
        return lambda: self._load_state_dict_post_hooks.remove(hook)

    def reset_parameters(self) -> None:
        for param in self.parameters():
            param.zero()

    def forward(self) -> Tensor | tuple[Tensor, ...]:
        raise NotImplementedError(
            "The forward method must be implemented by the subclass."
        )

    def train(self, mode: bool = True) -> Self:
        self.training = mode
        for module in self._modules.values():
            module.train(mode)
        return self

    def eval(self) -> Self:
        return self.train(mode=False)

    def to(self, device: _DeviceType) -> Self:
        if device == self.device:
            return self
        self.device = device

        for param in self.parameters(recurse=False):
            param.to(device)

        for buffer in self.buffers(recurse=False):
            buffer.to(device)

        for module in self.modules():
            module.to(device)

        return self

    def parameters(self, recurse: bool = True) -> Iterator[nn.Parameter]:
        seen_param_ids: set[int] = set()

        def _iter_parameters(module: Self) -> Iterator[nn.Parameter]:
            for _, param in module._parameters.items():
                param_id = id(param)
                if param_id in seen_param_ids:
                    continue
                seen_param_ids.add(param_id)
                yield param

            if recurse:
                for child_module in module._modules.values():
                    yield from _iter_parameters(child_module)

        yield from _iter_parameters(self)

    def buffers(self, recurse: bool = True) -> Iterator[nn.Buffer]:
        for buffer in self._buffers.values():
            yield buffer
        if recurse:
            for module in self._modules.values():
                yield from module.buffers(recurse=recurse)

    def modules(self) -> Iterator[Self]:
        yield self
        for module in self._modules.values():
            yield from module.modules()

    def children(self: Self) -> Iterator[Self]:
        return iter(self._modules.values())

    def count_parameters(self, recurse: bool = True) -> int:
        total_params = sum(p.size for p in self.parameters(recurse=recurse))
        return total_params

    @property
    def parameter_size(self) -> int:
        return self.count_parameters(recurse=True)

    def apply(self, fn: Callable[[Self, Any], None]) -> Self:
        fn(self)
        for module in self._modules.values():
            module.apply(fn)
        return self

    def state_dict(
        self,
        destination: OrderedDict[str, Any] | None = None,
        prefix: str = "",
        keep_vars: bool = False,
    ) -> OrderedDict:
        for hook in self._state_dict_pre_hooks:
            hook(self, prefix, keep_vars)

        if destination is None:
            destination = OrderedDict()

        for name, param in self._parameters.items():
            destination[prefix + name] = param if keep_vars else param.numpy()

        for name, buffer in self._buffers.items():
            destination[prefix + name] = buffer if keep_vars else buffer.numpy()

        for name, module in self._modules.items():
            module.state_dict(
                destination=destination, prefix=prefix + name + ".", keep_vars=keep_vars
            )

        for key in list(destination.keys()):
            if key in self._state_dict_pass_attr:
                del destination[key]

        for hook in self._state_dict_hooks:
            hook(self, destination, prefix, keep_vars)

        return destination

    def load_state_dict(
        self,
        state_dict: OrderedDict,
        strict: bool = True,
        *,
        verbose: bool = False,
        progress_desc: str | None = None,
    ) -> None:
        for hook in self._load_state_dict_pre_hooks:
            hook(self, state_dict, strict)

        own_state = self.state_dict(keep_vars=True)

        missing_keys = set(own_state.keys()) - set(state_dict.keys())
        unexpected_keys = set(state_dict.keys()) - set(own_state.keys())

        if strict:
            msg = ""
            if missing_keys:
                msg += f"Missing keys in state_dict: {missing_keys}\n"
            if unexpected_keys:
                msg += f"Unexpected keys in state_dict: {unexpected_keys}\n"
            if msg:
                raise KeyError("Error(s) in loading state_dict:\n" + msg)

        keys_to_migrate = [k for k in state_dict.keys() if k in own_state]
        total_params = len(keys_to_migrate)
        migrated = 0

        use_progress = verbose and tqdm is not None and total_params > 0
        pbar = (
            tqdm(
                total=total_params,
                desc=(
                    progress_desc
                    if progress_desc is not None
                    else "Applying state-dict"
                ),
                unit="param",
                leave=False,
            )
            if use_progress
            else None
        )

        try:
            for key in keys_to_migrate:
                value = state_dict[key]
                attr = own_state[key]
                if isinstance(attr, (nn.Parameter, nn.Buffer)):
                    value_t = Tensor(value, device=self.device)
                    attr.data = value_t.data
                else:
                    setattr(self, key, value)

                migrated += 1
                if pbar is not None:
                    shape = getattr(value, "shape", None)
                    shape_str = str(tuple(shape)) if shape is not None else "-"
                    pbar.set_postfix_str(
                        f"params={migrated}/{total_params}, shape={shape_str}, key={key}"
                    )
                    pbar.update(1)
        finally:
            if pbar is not None:
                pbar.close()

        for hook in self._load_state_dict_post_hooks:
            hook(self, missing_keys, unexpected_keys, strict)

    def __call__(self, *args: Any, **kwargs: Any) -> Tensor | tuple[Tensor, ...]:
        for hook, with_kwargs in self._forward_pre_hooks:
            if with_kwargs:
                result = hook(self, args, kwargs)
                if result is not None:
                    args, kwargs = result
            else:
                result = hook(self, args)
                if result is not None:
                    args = result

        output = self.forward(*args, **kwargs)

        for hook, with_kwargs in self._forward_hooks:
            if with_kwargs:
                result = hook(self, args, kwargs, output)
            else:
                result = hook(self, args, output)
            if result is not None:
                output = result

        if isinstance(output, Tensor) and self._backward_hooks:
            for hook in self._backward_hooks:
                output.register_hook(hook)

        if self._full_backward_pre_hooks or self._full_backward_hooks:
            outputs = output if isinstance(output, tuple) else (output,)
            output_tensors = [out for out in outputs if isinstance(out, Tensor)]

            if output_tensors:
                grad_outputs: list[_NumPyArray | None] = [None] * len(output_tensors)
                called = False

                def _call_full_backward_hooks() -> None:
                    nonlocal called, grad_outputs
                    if called:
                        return
                    called = True

                    grad_output_tuple = tuple(grad_outputs)
                    for hook in self._full_backward_pre_hooks:
                        result = hook(self, grad_output_tuple)
                        if result is not None:
                            grad_output_tuple = result

                    grad_input_tuple = tuple(
                        arg.grad if isinstance(arg, Tensor) else None for arg in args
                    )
                    for hook in self._full_backward_hooks:
                        hook(self, grad_input_tuple, grad_output_tuple)

                for idx, out in enumerate(output_tensors):

                    def _make_hook(index: int) -> Callable:
                        def _hook(_, grad: _NumPyArray) -> None:
                            grad_outputs[index] = grad
                            if all(g is not None for g in grad_outputs):
                                _call_full_backward_hooks()

                        return _hook

                    out.register_hook(_make_hook(idx))

        return output

    def compile(self, **kwargs) -> JITModule:
        return JITModule(self, **kwargs)

    def __repr__(self) -> str:
        extra = self.extra_repr()
        child_lines = []
        for name, module in self._modules.items():
            mod_str = repr(module)
            mod_str = _add_indent(mod_str, 2)
            child_lines.append(f"({name}): {mod_str}")

        main_str = self._get_name() + "("
        if extra:
            main_str += extra

        if child_lines:
            if extra:
                main_str += "\n"
            main_str += "\n  " + "\n  ".join(child_lines) + "\n"

        main_str += ")"
        return main_str

    def extra_repr(self) -> str:
        exclude = {"training", "device"}
        attrs = []
        for name, value in vars(self).items():
            if name.startswith("_") or name in exclude:
                continue
            if (
                name in self._parameters
                or name in self._buffers
                or name in self._modules
            ):
                continue
            attrs.append(f"{name}={value}")
        return ", ".join(attrs)

    def _get_name(self) -> str:
        return self._alt_name or type(self).__name__


def _add_indent(s: str, num_spaces: int) -> str:
    lines = s.splitlines()
    if len(lines) <= 1:
        return s
    first = lines[0]
    indented = [" " * num_spaces + line for line in lines[1:]]
    return "\n".join([first, *indented])


T = TypeVar("T", bound=Type[Module])


def auto_repr(*attr_names: str) -> Callable[[T], T]:
    def wrapper(cls: T) -> T:
        def extra_repr(self: Module) -> str:
            parts = []
            for name in attr_names:
                val = getattr(self, name, None)
                parts.append(f"{name}={val}")
            return ", ".join(parts)

        cls.extra_repr = extra_repr
        return cls

    return wrapper


def set_state_dict_pass_attr(*attr_names: str) -> Callable[[T], T]:
    def wrapper(cls: T) -> T:
        cls._state_dict_pass_attr = set(attr_names)
        return cls

    return wrapper


class Sequential(Module):
    @overload
    def __init__(self, *modules: Module) -> None: ...

    @overload
    def __init__(self, ordered_dict: OrderedDict[str, Module]) -> None: ...

    def __init__(self, *args: Module | OrderedDict[str, Module]) -> None:
        super().__init__()
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for name, module in args[0].items():
                self.add_module(name, module)
        else:
            for idx, module in enumerate(args):
                self.add_module(str(idx), module)

    def forward(self, input: Tensor) -> Tensor:
        for module in self._modules.values():
            input = module(input)
        return input

    def __getitem__(self, idx: int | slice) -> Module | Self:
        if isinstance(idx, slice):
            modules_slice = list(self._modules.items())[idx]
            return Sequential(OrderedDict(modules_slice))

        elif isinstance(idx, int):
            if idx < 0:
                idx += len(self._modules)
            keys = list(self._modules.keys())

            if idx < 0 or idx >= len(keys):
                raise IndexError("Index out of range")

            return self._modules[keys[idx]]
        else:
            raise TypeError(f"Invalid index type: {type(idx)}. Must be int or slice.")

    def __setitem__(self, idx: int, module: Module) -> None:
        if not isinstance(idx, int):
            raise TypeError("Indices should be integers for __setitem__.")

        keys = list(self._modules.keys())
        if idx < 0:
            idx += len(keys)
        if idx < 0 or idx >= len(keys):
            raise IndexError("Index out of range")

        old_key = keys[idx]
        del self._modules[old_key]
        self._modules[old_key] = module

    def __delitem__(self, idx: int) -> None:
        if not isinstance(idx, int):
            raise TypeError("Indices should be integers for __delitem__.")

        keys = list(self._modules.keys())
        if idx < 0:
            idx += len(keys)
        if idx < 0 or idx >= len(keys):
            raise IndexError("Index out of range")

        del self._modules[keys[idx]]

    def __len__(self) -> int:
        return len(self._modules)

    def append(self, module: Module) -> None:
        self.add_module(str(len(self._modules)), module)

    def extend(self, modules: Iterator[Module]) -> None:
        for module in modules:
            self.append(module)

    @classmethod
    def from_ordered_dict(cls: type[Self], odict: OrderedDict[str, Module]) -> Self:
        return cls(odict)

    @classmethod
    def from_modules(cls: type[Self], *modules: Module) -> Self:
        return cls(*modules)


class ModuleList(Module):
    def __init__(self, modules: list[Module] | None = None) -> None:
        super().__init__()
        if modules is not None:
            self.extend(modules)

    def __getitem__(self, idx: int | slice) -> Module | Self:
        if isinstance(idx, slice):
            mod_slice = list(self._modules.items())[idx]
            ml = ModuleList()
            for i, (_, m) in enumerate(mod_slice):
                ml.add_module(str(i), m)
            return ml

        elif isinstance(idx, int):
            if idx < 0:
                idx += len(self._modules)
            keys = list(self._modules.keys())
            if idx < 0 or idx >= len(keys):
                raise IndexError("Index out of range.")

            return self._modules[keys[idx]]
        else:
            raise TypeError(f"Invalid index type: {type(idx)}. Must be int or slice.")

    def __setitem__(self, idx: int, module: Module) -> None:
        if not isinstance(idx, int):
            raise TypeError("Indices should be integers.")

        keys = list(self._modules.keys())
        if idx < 0:
            idx += len(keys)
        if idx < 0 or idx >= len(keys):
            raise IndexError("Index out of range.")

        old_key = keys[idx]
        del self._modules[old_key]
        self.add_module(old_key, module)

    def __delitem__(self, idx: int) -> None:
        if not isinstance(idx, int):
            raise TypeError("Indices should be integers.")

        keys = list(self._modules.keys())
        if idx < 0:
            idx += len(keys)
        if idx < 0 or idx >= len(keys):
            raise IndexError("Index out of range.")

        items = list(self._modules.items())
        self._modules.clear()
        for i, (_, m) in enumerate(items):
            self._modules[str(i)] = m

    def __len__(self) -> int:
        return len(self._modules)

    def __iter__(self) -> Iterator[Module]:
        return iter(self._modules.values())

    def append(self, module: Module) -> None:
        self.add_module(str(len(self._modules)), module)

    def extend(self, modules: list[Module]) -> None:
        for m in modules:
            self.append(m)

    def insert(self, index: int, module: Module) -> None:
        if not isinstance(index, int):
            raise TypeError("Index should be an integer.")

        total = len(self._modules)
        if index < 0:
            index += total
        if index < 0:
            index = 0
        if index > total:
            index = total

        items = list(self._modules.items())
        items.insert(index, (str(index), module))

        self._modules.clear()
        for i, (_, m) in enumerate(items):
            self._modules[str(i)] = m


class ModuleDict(Module):
    def __init__(self, modules: dict[str, Module] | None = None) -> None:
        super().__init__()
        if modules is not None:
            self.update(modules)

    def update(self, modules: dict[str, Module]) -> None:
        for k, m in modules.items():
            self[k] = m

    def clear(self) -> None:
        self._modules.clear()

    def pop(self, key: str) -> Module:
        module = self._modules[key]
        del self._modules[key]
        return module

    def __getitem__(self, key: str) -> Module:
        return self._modules[key]

    def __setitem__(self, key: str, module: Module) -> None:
        if not isinstance(module, Module):
            raise TypeError(f"Expected Module, got {type(module)}.")
        self.add_module(key, module)

    def __delitem__(self, key: str) -> None:
        del self._modules[key]

    def __len__(self) -> int:
        return len(self._modules)

    def __iter__(self) -> Iterator[str]:
        return iter(self._modules)

    def keys(self) -> KeysView[str]:
        return self._modules.keys()

    def values(self) -> ValuesView[Module]:
        return self._modules.values()

    def items(self) -> ItemsView[str, Module]:
        return self._modules.items()


class ParameterList(Module):
    def __init__(self, parameters: list[nn.Parameter] | None = None) -> None:
        super().__init__()
        if parameters is not None:
            self.extend(parameters)

    def __getitem__(self, idx: int | slice) -> nn.Parameter | Self:
        if isinstance(idx, slice):
            items = list(self._parameters.items())[idx]
            plist = ParameterList()

            for i, (_, p) in enumerate(items):
                plist.register_parameter(str(i), p)
            return plist

        elif isinstance(idx, int):
            if idx < 0:
                idx += len(self._parameters)
            keys = list(self._parameters.keys())

            if idx < 0 or idx >= len(keys):
                raise IndexError("Index out of range.")
            return self._parameters[keys[idx]]

        else:
            return TypeError(f"Invalid index type: {type(idx)}. Must be int or slice.")

    def __setitem__(self, idx: int, param: nn.Parameter) -> None:
        if not isinstance(idx, int):
            raise TypeError("Indices should be integers.")
        if not isinstance(param, nn.Parameter):
            raise TypeError("Can only set Parameter in ParameterList.")

        keys = list(self._parameters.keys())
        if idx < 0:
            idx += len(keys)
        if idx < 0 or idx >= len(keys):
            raise IndexError("Index out of range.")

        old_key = keys[idx]
        del self._parameters[old_key]
        self.register_parameter(old_key, param)

    def __delitem__(self, idx: int) -> None:
        if not isinstance(idx, int):
            raise TypeError("Indices should be integers.")

        keys = list(self._parameters.keys())
        if idx < 0:
            idx += len(keys)
        if idx < 0 or idx >= len(keys):
            raise IndexError("Index out of range")

        del self._parameters[keys[idx]]

        items = list(self._parameters.items())
        self._parameters.clear()
        for i, (_, p) in enumerate(items):
            self._parameters[str(i)] = p

    def __len__(self) -> int:
        return len(self._parameters)

    def __iter__(self) -> Iterator[nn.Parameter]:
        return iter(self._parameters.values())

    def append(self, param: nn.Parameter) -> None:
        if not isinstance(param, nn.Parameter):
            raise TypeError("Can only append Parameter to ParameterList.")
        self.register_parameter(str(len(self._parameters)), param)

    def extend(self, parameters) -> None:
        for p in parameters:
            self.append(p)

    def insert(self, index: int, param: nn.Parameter) -> None:
        if not isinstance(index, int):
            raise TypeError("Index should be an integer for insert.")
        if not isinstance(param, nn.Parameter):
            raise TypeError("Can only insert Parameter into ParameterList.")

        total = len(self._parameters)
        if index < 0:
            index += total
        if index < 0:
            index = 0
        if index > total:
            index = total

        items = list(self._parameters.items())
        items.insert(index, (str(index), param))

        self._parameters.clear()
        for i, (_, p) in enumerate(items):
            self._parameters[str(i)] = p

    def __repr__(self) -> str:
        lines = []
        for i, (_, p) in enumerate(self._parameters.items()):
            param_str = _add_indent(repr(p), 2)
            lines.append(f"({i}): {param_str}")

        main_str = self._get_name() + "("
        if lines:
            main_str += "\n  " + "\n  ".join(lines) + "\n"
        main_str += ")"
        return main_str


class ParameterDict(Module):
    def __init__(self, parameters: dict[str, nn.Parameter] | None = None) -> None:
        super().__init__()
        if parameters is not None:
            self.update(parameters)

    def update(self, parameters: dict[str, nn.Parameter]) -> None:
        for k, p in parameters.items():
            self[k] = p

    def clear(self) -> None:
        self._parameters.clear()

    def pop(self, key: str) -> nn.Parameter:
        param = self._parameters[key]
        del self._parameters[key]
        return param

    def __getitem__(self, key: str) -> nn.Parameter:
        return self._parameters[key]

    def __setitem__(self, key: str, param: nn.Parameter) -> None:
        if not isinstance(param, nn.Parameter):
            raise TypeError(f"Expected nn.Parameter, got {type(param)}")
        self.register_parameter(key, param)

    def __delitem__(self, key: str) -> None:
        del self._parameters[key]

    def __len__(self) -> int:
        return len(self._parameters)

    def __iter__(self) -> Iterator[str]:
        return iter(self._parameters)

    def keys(self) -> KeysView[str]:
        return self._parameters.keys()

    def values(self) -> ValuesView[nn.Parameter]:
        return self._parameters.values()

    def items(self) -> ItemsView[str, nn.Parameter]:
        return self._parameters.items()

    def __repr__(self) -> str:
        lines = []
        for name, param in self._parameters.items():
            param_str = _add_indent(repr(param), 2)
            lines.append(f"({name}): {param_str}")

        main_str = self._get_name() + "("
        if lines:
            main_str += "\n  " + "\n  ".join(lines) + "\n"
        main_str += ")"
        return main_str



from typing import Any

from lucid._tensor.tensor import Tensor as _Tensor


def _is_namedtuple(value: Any) -> bool:
    t = type(value)
    return (
        isinstance(value, tuple)
        and hasattr(t, "_fields")
        and isinstance(t._fields, tuple)
    )


def flatten_output(value: Any) -> tuple[list[_Tensor], Any]:
    if isinstance(value, _Tensor):
        return [value], ("leaf",)

    if _is_namedtuple(value):
        nt_type = type(value)
        flat: list[_Tensor] = []
        child_specs: list[Any] = []
        for item in value:
            child_flat, child_spec = flatten_output(item)
            flat.extend(child_flat)
            child_specs.append(child_spec)
        return flat, ("namedtuple", nt_type, child_specs)

    if isinstance(value, tuple):
        flat = []
        child_specs = []
        for item in value:
            child_flat, child_spec = flatten_output(item)
            flat.extend(child_flat)
            child_specs.append(child_spec)
        return flat, ("tuple", child_specs)

    if isinstance(value, list):
        flat = []
        child_specs = []
        for item in value:
            child_flat, child_spec = flatten_output(item)
            flat.extend(child_flat)
            child_specs.append(child_spec)
        return flat, ("list", child_specs)

    if isinstance(value, dict):
        keys = list(value.keys())
        flat = []
        child_specs = []
        for k in keys:
            child_flat, child_spec = flatten_output(value[k])
            flat.extend(child_flat)
            child_specs.append(child_spec)
        return flat, ("dict", keys, child_specs)

    return [], ("const", value)


def unflatten_output(flat: list[_Tensor], treespec: Any) -> Any:
    _, result = _unflatten(flat, treespec, 0)
    return result


def _unflatten(flat: list[_Tensor], treespec: Any, offset: int) -> tuple[int, Any]:
    kind = treespec[0]

    if kind == "leaf":
        return offset + 1, flat[offset]

    if kind == "const":
        return offset, treespec[1]

    if kind in ("tuple", "list"):
        _, child_specs = treespec
        items = []
        for cs in child_specs:
            offset, item = _unflatten(flat, cs, offset)
            items.append(item)
        result = tuple(items) if kind == "tuple" else items
        return offset, result

    if kind == "dict":
        _, keys, child_specs = treespec
        result = {}
        for k, cs in zip(keys, child_specs):
            offset, val = _unflatten(flat, cs, offset)
            result[k] = val
        return offset, result

    if kind == "namedtuple":
        _, nt_type, child_specs = treespec
        items = []
        for cs in child_specs:
            offset, item = _unflatten(flat, cs, offset)
            items.append(item)
        return offset, nt_type(*items)

    raise ValueError(f"Unknown treespec kind: {kind!r}")



from lucid.nn.parameter import *
from lucid.nn.module import *
from lucid.nn.modules import *
from lucid.nn.fused import *
from lucid.nn.cache import *

import lucid.nn.init as init
import lucid.nn.utils as utils



from lucid.nn._kernel.attention import scaled_dot_product_attention_kernel

from lucid._tensor import Tensor


def scaled_dot_product_attention(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    attn_mask: Tensor | None = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    scale: float | None = None,
    output_weight: bool = False,
) -> Tensor | tuple[Tensor, Tensor]:
    op = scaled_dot_product_attention_kernel(
        attn_mask=attn_mask,
        is_causal=is_causal,
        scale=scale,
        dropout_p=dropout_p,
    )
    attn_output = op(query, key, value)
    if not output_weight:
        return attn_output
    return attn_output, op.get_attention_weight(device=query.device)



import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.base import PreTrainedModelMixin


from ._model import GPT, GPTConfig

__all__ = ["GPTLMHeadModel", "GPTDoubleHeadsModel", "GPTForSequenceClassification"]


class _GPTTaskWrapperMixin:
    gpt: GPT

    def get_input_embeddings(self) -> nn.Embedding:
        return self.gpt.get_input_embeddings()

    def set_input_embeddings(self, value: nn.Embedding) -> None:
        self.gpt.set_input_embeddings(value)

    def get_output_embeddings(self) -> nn.Linear | None:
        return getattr(self, "lm_head", None)

    def set_output_embeddings(self, new_embeddings: nn.Linear) -> None:
        self.lm_head = new_embeddings

    def tie_weights(self) -> None:
        out = self.get_output_embeddings()
        if out is not None:
            out.weight = self.get_input_embeddings().weight


def _causal_lm_loss(logits: Tensor, labels: Tensor) -> Tensor:
    shift_logits = logits[:, :-1]
    shift_labels = labels[:, 1:]
    return F.cross_entropy(
        shift_logits.reshape(-1, shift_logits.shape[-1]),
        shift_labels.reshape(-1),
        ignore_index=-100,
    )


def _classification_ce_loss(logits: Tensor, labels: Tensor) -> Tensor:
    return F.cross_entropy(logits.reshape(-1, logits.shape[-1]), labels.reshape(-1))


class GPTLMHeadModel(_GPTTaskWrapperMixin, PreTrainedModelMixin, nn.Module):
    def __init__(self, config: GPTConfig) -> None:
        super().__init__()
        self.config = config
        self.gpt = GPT(config)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.tie_weights()

    def forward(
        self,
        input_ids: Tensor,
        attention_mask: Tensor | None = None,
        position_ids: Tensor | None = None,
        past_key_values: list[nn.KVCache] | None = None,
        labels: Tensor | None = None,
        use_cache: bool = False,
    ) -> tuple[Tensor | None, Tensor, list[nn.KVCache] | None]:
        hidden_states, presents = self.gpt(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            use_cache=use_cache,
        )
        logits = self.lm_head(hidden_states)

        loss = _causal_lm_loss(logits, labels) if labels is not None else None
        return loss, logits, presents


class _GPTMultipleChoiceHead(nn.Module):
    def __init__(self, config: GPTConfig) -> None:
        super().__init__()
        self.summary = nn.Linear(config.hidden_size, 1)
        self.drop = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states: Tensor, mc_token_ids: Tensor) -> Tensor:
        N, C, L, H = hidden_states.shape
        idx = mc_token_ids.unsqueeze(axis=-1).unsqueeze(axis=-1).expand(N, C, 1, H)
        mc_hidden = hidden_states.gather(axis=2, index=idx).squeeze(axis=2)

        return self.summary(self.drop(mc_hidden)).squeeze(axis=-1)


class GPTDoubleHeadsModel(_GPTTaskWrapperMixin, PreTrainedModelMixin, nn.Module):
    def __init__(self, config: GPTConfig) -> None:
        super().__init__()
        self.config = config
        self.gpt = GPT(config)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.mc_head = _GPTMultipleChoiceHead(config)
        self.tie_weights()

    def forward(
        self,
        input_ids: Tensor,
        mc_token_ids: Tensor,
        attention_mask: Tensor | None = None,
        position_ids: Tensor | None = None,
        labels: Tensor | None = None,
        mc_labels: Tensor | None = None,
        use_cache: bool = False,
    ) -> tuple[Tensor | None, Tensor | None, Tensor, Tensor, Tensor]:
        N, C, L = input_ids.shape

        flat_input_ids = input_ids.reshape(N * C, L)
        flat_attn_mask = (
            attention_mask.reshape(N * C, L) if attention_mask is not None else None
        )
        flat_position_ids = (
            position_ids.reshape(N * C, L) if position_ids is not None else None
        )

        hidden_states, presents = self.gpt(
            flat_input_ids,
            attention_mask=flat_attn_mask,
            position_ids=flat_position_ids,
            use_cache=use_cache,
        )

        lm_logits = self.lm_head(hidden_states).reshape(N, C, L, -1)
        mc_logits = self.mc_head(hidden_states.reshape(N, C, L, -1), mc_token_ids)

        lm_loss, mc_loss = None, None
        if labels is not None:
            flat_logits = lm_logits.reshape(N * C, L, -1)
            flat_labels = labels.reshape(N * C, L)
            lm_loss = _causal_lm_loss(flat_logits, flat_labels)

        if mc_labels is not None:
            mc_loss = F.cross_entropy(mc_logits, mc_labels)

        return lm_loss, mc_loss, lm_logits, mc_logits, presents


class GPTForSequenceClassification(
    _GPTTaskWrapperMixin, PreTrainedModelMixin, nn.Module
):
    def __init__(self, config: GPTConfig, num_labels: int = 2) -> None:
        super().__init__()
        self.config = config
        self.num_labels = num_labels

        self.gpt = GPT(config)
        self.score = nn.Linear(config.hidden_size, num_labels, bias=False)
        self.apply(self.gpt._init_weights)

    def forward(
        self,
        input_ids: Tensor,
        attention_mask: Tensor | None = None,
        position_ids: Tensor | None = None,
        labels: Tensor | None = None,
    ) -> tuple[Tensor | None, Tensor]:
        hidden_states, _ = self.gpt(
            input_ids, attention_mask=attention_mask, position_ids=position_ids
        )
        if self.config.pad_token_id is None:
            seq_lengths = -1
        else:
            seq_lengths = (input_ids != self.config.pad_token_id).sum(axis=-1) - 1

        pooled = hidden_states[range(hidden_states.shape[0]), seq_lengths]
        logits = self.score(pooled)

        loss = _classification_ce_loss(logits, labels) if labels is not None else None
        return loss, logits



import lucid
import lucid.nn.functional

from lucid._tensor import Tensor
from lucid.types import _Scalar, Numeric


def _interpolate_bilinear(
    input_: Tensor, size: tuple[int, int], align_corners: bool = False
) -> Tensor:
    _, _, H, W = input_.shape
    out_h, out_w = size
    if (H, W) == (out_h, out_w):
        return input_

    device = input_.device
    if align_corners:
        if out_h == 1:
            indices_h = lucid.zeros((out_h,), dtype=lucid.Float32, device=device)
        else:
            indices_h = (
                lucid.arange(out_h, dtype=lucid.Float32, device=device)
                * (H - 1)
                / (out_h - 1)
            )
        if out_w == 1:
            indices_w = lucid.zeros((out_w,), dtype=lucid.Float32, device=device)
        else:
            indices_w = (
                lucid.arange(out_w, dtype=lucid.Float32, device=device)
                * (W - 1)
                / (out_w - 1)
            )
    else:
        scale_h = H / out_h
        scale_w = W / out_w
        indices_h = (
            lucid.arange(out_h, dtype=lucid.Float32, device=device) + 0.5
        ) * scale_h - 0.5
        indices_w = (
            lucid.arange(out_w, dtype=lucid.Float32, device=device) + 0.5
        ) * scale_w - 0.5

    indices_h = indices_h.clip(0, H - 1)
    indices_w = indices_w.clip(0, W - 1)

    top_indices_f = lucid.floor(indices_h)
    left_indices_f = lucid.floor(indices_w)

    top_indices = top_indices_f.astype(lucid.Int)
    bot_indices = (top_indices_f + 1).clip(0, H - 1).astype(lucid.Int)
    left_indices = left_indices_f.astype(lucid.Int)
    right_indices = (left_indices_f + 1).clip(0, W - 1).astype(lucid.Int)

    h_lerp = indices_h - top_indices_f
    w_lerp = indices_w - left_indices_f

    top_left = input_[:, :, top_indices[:, None], left_indices]
    top_right = input_[:, :, top_indices[:, None], right_indices]
    bot_left = input_[:, :, bot_indices[:, None], left_indices]
    bot_right = input_[:, :, bot_indices[:, None], right_indices]

    top = top_left * (1 - w_lerp) + top_right * w_lerp
    bot = bot_left * (1 - w_lerp) + bot_right * w_lerp

    interpolated = top * (1 - h_lerp[:, None]) + bot * h_lerp[:, None]
    return interpolated


def _interpolate_trilinear(
    input_: Tensor, size: tuple[int, int, int], align_corners: bool = False
) -> Tensor:
    _, _, D, H, W = input_.shape
    out_d, out_h, out_w = size
    if (D, H, W) == (out_d, out_h, out_w):
        return input_

    device = input_.device
    if align_corners:
        indices_d = (
            lucid.zeros((out_d,), dtype=lucid.Float32, device=device)
            if out_d == 1
            else lucid.arange(out_d, dtype=lucid.Float32, device=device)
            * (D - 1)
            / (out_d - 1)
        )
        indices_h = (
            lucid.zeros((out_h,), dtype=lucid.Float32, device=device)
            if out_h == 1
            else lucid.arange(out_h, dtype=lucid.Float32, device=device)
            * (H - 1)
            / (out_h - 1)
        )
        indices_w = (
            lucid.zeros((out_w,), dtype=lucid.Float32, device=device)
            if out_w == 1
            else lucid.arange(out_w, dtype=lucid.Float32, device=device)
            * (W - 1)
            / (out_w - 1)
        )
    else:
        indices_d = (lucid.arange(out_d, dtype=lucid.Float32, device=device) + 0.5) * (
            D / out_d
        ) - 0.5
        indices_h = (lucid.arange(out_h, dtype=lucid.Float32, device=device) + 0.5) * (
            H / out_h
        ) - 0.5
        indices_w = (lucid.arange(out_w, dtype=lucid.Float32, device=device) + 0.5) * (
            W / out_w
        ) - 0.5

    indices_d = indices_d.clip(0, D - 1)
    indices_h = indices_h.clip(0, H - 1)
    indices_w = indices_w.clip(0, W - 1)

    d0_f = lucid.floor(indices_d)
    h0_f = lucid.floor(indices_h)
    w0_f = lucid.floor(indices_w)

    d0 = d0_f.astype(lucid.Int)
    d1 = (d0_f + 1).clip(0, D - 1).astype(lucid.Int)
    h0 = h0_f.astype(lucid.Int)
    h1 = (h0_f + 1).clip(0, H - 1).astype(lucid.Int)
    w0 = w0_f.astype(lucid.Int)
    w1 = (w0_f + 1).clip(0, W - 1).astype(lucid.Int)

    d_lerp = indices_d - d0_f
    h_lerp = indices_h - h0_f
    w_lerp = indices_w - w0_f

    c000 = input_[:, :, d0[:, None, None], h0[None, :, None], w0[None, None, :]]
    c001 = input_[:, :, d0[:, None, None], h0[None, :, None], w1[None, None, :]]

    c010 = input_[:, :, d0[:, None, None], h1[None, :, None], w0[None, None, :]]
    c011 = input_[:, :, d0[:, None, None], h1[None, :, None], w1[None, None, :]]

    c100 = input_[:, :, d1[:, None, None], h0[None, :, None], w0[None, None, :]]
    c101 = input_[:, :, d1[:, None, None], h0[None, :, None], w1[None, None, :]]

    c110 = input_[:, :, d1[:, None, None], h1[None, :, None], w0[None, None, :]]
    c111 = input_[:, :, d1[:, None, None], h1[None, :, None], w1[None, None, :]]

    c00 = c000 * (1 - w_lerp) + c001 * w_lerp
    c01 = c010 * (1 - w_lerp) + c011 * w_lerp
    c10 = c100 * (1 - w_lerp) + c101 * w_lerp
    c11 = c110 * (1 - w_lerp) + c111 * w_lerp

    c0 = c00 * (1 - h_lerp[:, None]) + c01 * h_lerp[:, None]
    c1 = c10 * (1 - h_lerp[:, None]) + c11 * h_lerp[:, None]

    return c0 * (1 - d_lerp[:, None, None]) + c1 * d_lerp[:, None, None]


def _interpolate_nearest_3d(
    input_: Tensor, size: tuple[int, int, int], align_corners: bool = False
) -> Tensor:
    _ = align_corners
    _, _, D, H, W = input_.shape
    device = input_.device
    out_d, out_h, out_w = size

    if (D, H, W) == (out_d, out_h, out_w):
        return input_

    indices_d = (
        lucid.floor(
            lucid.arange(out_d, dtype=lucid.Float32, device=device) * (D / out_d)
        )
        .clip(0, D - 1)
        .astype(lucid.Int32)
    )
    indices_h = (
        lucid.floor(
            lucid.arange(out_h, dtype=lucid.Float32, device=device) * (H / out_h)
        )
        .clip(0, H - 1)
        .astype(lucid.Int32)
    )
    indices_w = (
        lucid.floor(
            lucid.arange(out_w, dtype=lucid.Float32, device=device) * (W / out_w)
        )
        .clip(0, W - 1)
        .astype(lucid.Int32)
    )

    return input_[
        :,
        :,
        indices_d[:, None, None],
        indices_h[None, :, None],
        indices_w[None, None, :],
    ]


def _interpolate_nearest(
    input_: Tensor, size: tuple[int, int], align_corners: bool = False
) -> Tensor:
    _ = align_corners
    _, _, H, W = input_.shape
    device = input_.device
    out_h, out_w = size

    if (H, W) == (out_h, out_w):
        return input_

    scale_h = H / out_h
    scale_w = W / out_w

    indices_h = lucid.floor(
        lucid.arange(out_h, dtype=lucid.Float32, device=device) * scale_h
    )
    indices_w = lucid.floor(
        lucid.arange(out_w, dtype=lucid.Float32, device=device) * scale_w
    )
    indices_h = indices_h.clip(0, H - 1).astype(lucid.Int32)
    indices_w = indices_w.clip(0, W - 1).astype(lucid.Int32)

    return input_[:, :, indices_h[:, None], indices_w[None, :]]


def _interpolate_area(
    input_: Tensor, size: tuple[int, int], align_corners: bool = False
) -> Tensor:
    _, _, H, W = input_.shape
    out_h, out_w = size

    scale_h = H / out_h
    scale_w = W / out_w

    pooled = lucid.nn.functional.avg_pool2d(
        input_,
        kernel_size=(int(scale_h), int(scale_w)),
        stride=(int(scale_h), int(scale_w)),
    )
    return pooled[:, :, out_h, out_w]


def rotate(
    input_: Tensor, angle: float, center: tuple[_Scalar, _Scalar] | None = None
) -> Tensor:
    N, C, H, W = input_.shape

    if center is None:
        center_x = W / 2
        center_y = H / 2
    else:
        center_x, center_y = center

    angle_rad = -angle * (lucid.pi / 180)
    cos_a = lucid.cos(angle_rad).data
    sin_a = lucid.sin(angle_rad).data

    rot_mat = [
        [cos_a, -sin_a, center_x - cos_a * center_x + sin_a * center_y],
        [sin_a, cos_a, center_y - sin_a * center_x - cos_a * center_y],
    ]
    rot_mat = lucid.to_tensor(rot_mat)

    y_coords, x_coords = lucid.arange(H), lucid.arange(W)
    y_grid, x_grid = lucid.meshgrid(y_coords, x_coords, indexing="ij")

    x_flat = x_grid.ravel()
    y_flat = y_grid.ravel()

    ones = lucid.ones_like(x_flat)
    homogen_coords = lucid.stack([x_flat, y_flat, ones])

    new_coords = (rot_mat @ homogen_coords).free()
    new_x = new_coords[0].reshape(H, W)
    new_y = new_coords[1].reshape(H, W)

    new_x = new_x.clip(0, W - 1).astype(lucid.Int)
    new_y = new_y.clip(0, H - 1).astype(lucid.Int)

    rotated_img = lucid.zeros_like(input_, device=input_.device)
    for n in range(N):
        for c in range(C):
            rotated_img[n, c] = input_[n, c, new_y, new_x]

    return rotated_img


def one_hot(
    input_: Tensor, num_classes: int = -1, dtype: Numeric | bool | None = None
) -> Tensor:
    if input_.dtype.base_dtype is not int:
        raise TypeError("one_hot only supports integer input.")
    if num_classes == -1:
        num_classes = lucid.max(input_).item() + 1

    input_flat = input_.reshape(-1)
    N = input_flat.shape[0]

    out_shape = (*input_.shape, num_classes)
    out = lucid.zeros(N, num_classes, device=input_.device, dtype=lucid.Int8)

    arange = lucid.arange(N, device=input_.device, dtype=lucid.Int32)
    out[arange, input_flat.astype(lucid.Int)] = 1

    return (
        out.reshape(out_shape).astype(dtype)
        if dtype is not None
        else out.reshape(out_shape)
    )



from dataclasses import dataclass

import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor
from lucid.models.base import PreTrainedModelMixin

__all__ = ["VGGNet", "VGGNetConfig", "vggnet_11", "vggnet_13", "vggnet_16", "vggnet_19"]


@dataclass
class VGGNetConfig:
    conv_config: list[int | str]
    num_classes: int = 1000
    in_channels: int = 3
    dropout: float = 0.5
    classifier_hidden_features: tuple[int, int] = (4096, 4096)

    def __post_init__(self) -> None:
        if not self.conv_config:
            raise ValueError("VGGNet conv_config must not be empty.")

        if self.num_classes <= 0:
            raise ValueError("VGGNet num_classes must be greater than 0.")

        if self.in_channels <= 0:
            raise ValueError("VGGNet in_channels must be greater than 0.")

        if not 0.0 <= self.dropout < 1.0:
            raise ValueError("VGGNet dropout must be in the range [0.0, 1.0).")

        if len(self.classifier_hidden_features) != 2:
            raise ValueError(
                "VGGNet classifier_hidden_features must contain exactly 2 values."
            )

        if any(hidden_dim <= 0 for hidden_dim in self.classifier_hidden_features):
            raise ValueError(
                "VGGNet classifier_hidden_features values must be greater than 0."
            )

        saw_conv = False
        for idx, layer in enumerate(self.conv_config, start=1):
            if layer == "M":
                continue
            if not isinstance(layer, int) or layer <= 0:
                raise ValueError(
                    "VGGNet conv_config entries must be positive integers or 'M' "
                    f"(got {layer!r} at index {idx})."
                )
            saw_conv = True

        if not saw_conv:
            raise ValueError("VGGNet conv_config must include at least 1 conv layer.")


class VGGNet(nn.Module, PreTrainedModelMixin):
    def __init__(self, config: VGGNetConfig) -> None:
        super().__init__()
        self.config = config
        self.conv = self._make_layers(config)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(7, 7))
        hidden_1, hidden_2 = config.classifier_hidden_features
        out_channels = next(
            layer for layer in reversed(config.conv_config) if isinstance(layer, int)
        )
        self.fc = nn.Sequential(
            nn.Linear(out_channels * 7 * 7, hidden_1),
            nn.ReLU(),
            nn.Dropout(config.dropout),
            nn.Linear(hidden_1, hidden_2),
            nn.ReLU(),
            nn.Dropout(config.dropout),
            nn.Linear(hidden_2, config.num_classes),
        )

    def _make_layers(self, config: VGGNetConfig) -> nn.Sequential:
        layers = []
        in_channels = config.in_channels
        for layer in config.conv_config:
            if layer == "M":
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))
            else:
                layers.append(nn.Conv2d(in_channels, layer, kernel_size=3, padding=1))
                layers.append(nn.ReLU())

                in_channels = layer

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)
        return x


def _vgg_config_from_kwargs(
    *,
    conv_config: list[int | str],
    num_classes: int,
    **kwargs,
) -> VGGNetConfig:
    return VGGNetConfig(conv_config=conv_config, num_classes=num_classes, **kwargs)


@register_model
def vggnet_11(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, "M"]
    config.extend([128, "M"])
    config.extend([256, 256, "M"])
    config.extend([512, 512, "M", 512, 512, "M"])

    model_config = _vgg_config_from_kwargs(
        conv_config=config,
        num_classes=num_classes,
        **kwargs,
    )
    return VGGNet(model_config)


@register_model
def vggnet_13(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, 64, "M"]
    config.extend([128, 128, "M"])
    config.extend([256, 256, "M"])
    config.extend([512, 512, "M", 512, 512, "M"])

    model_config = _vgg_config_from_kwargs(
        conv_config=config,
        num_classes=num_classes,
        **kwargs,
    )
    return VGGNet(model_config)


@register_model
def vggnet_16(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, 64, "M"]
    config.extend([128, 128, "M"])
    config.extend([256, 256, 256, "M"])
    config.extend([512, 512, 512, "M", 512, 512, 512, "M"])

    model_config = _vgg_config_from_kwargs(
        conv_config=config,
        num_classes=num_classes,
        **kwargs,
    )
    return VGGNet(model_config)


@register_model
def vggnet_19(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, 64, "M"]
    config.extend([128, 128, "M"])
    config.extend([256, 256, 256, 256, "M"])
    config.extend([512, 512, 512, 512, "M", 512, 512, 512, 512, "M"])

    model_config = _vgg_config_from_kwargs(
        conv_config=config,
        num_classes=num_classes,
        **kwargs,
    )
    return VGGNet(model_config)



from dataclasses import dataclass
from typing import Type

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

__all__ = [
    "CoAtNet",
    "CoAtNetConfig",
    "coatnet_0",
    "coatnet_1",
    "coatnet_2",
    "coatnet_3",
    "coatnet_4",
    "coatnet_5",
    "coatnet_6",
    "coatnet_7",
]


def _normalize_positive_int_sequence(
    values: tuple[int, ...] | list[int],
    name: str,
    *,
    expected_length: int | None = None,
) -> tuple[int, ...]:
    normalized = tuple(values)
    if expected_length is not None and len(normalized) != expected_length:
        raise ValueError(f"{name} must contain exactly {expected_length} values")
    if len(normalized) == 0:
        raise ValueError(f"{name} must contain at least one value")
    if any(not isinstance(value, int) or value <= 0 for value in normalized):
        raise ValueError(f"{name} values must be positive integers")
    return normalized


@dataclass
class CoAtNetConfig:
    img_size: tuple[int, int] | list[int]
    in_channels: int
    num_blocks: tuple[int, ...] | list[int]
    channels: tuple[int, ...] | list[int]
    num_classes: int = 1000
    num_heads: int = 32
    block_types: tuple[str, ...] | list[str] = ("C", "C", "T", "T")
    scaled_num_blocks: tuple[int, int] | list[int] | None = None
    scaled_channels: tuple[int, int] | list[int] | None = None

    def __post_init__(self) -> None:
        if not isinstance(self.img_size, (list, tuple)) or len(self.img_size) != 2:
            raise ValueError("img_size must contain exactly two values")
        self.img_size = tuple(self.img_size)
        if any(not isinstance(value, int) or value < 32 for value in self.img_size):
            raise ValueError(
                "img_size values must be integers greater than or equal to 32"
            )

        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if self.num_heads <= 0:
            raise ValueError("num_heads must be greater than 0")

        self.num_blocks = _normalize_positive_int_sequence(
            self.num_blocks, "num_blocks", expected_length=5
        )
        self.channels = _normalize_positive_int_sequence(
            self.channels, "channels", expected_length=5
        )

        if not isinstance(self.block_types, (list, tuple)):
            raise TypeError("block_types must be a sequence")
        self.block_types = tuple(self.block_types)
        if len(self.block_types) != 4:
            raise ValueError("block_types must contain exactly 4 values")
        if any(block_type not in {"C", "T"} for block_type in self.block_types):
            raise ValueError("block_types values must be 'C' or 'T'")

        if (self.scaled_num_blocks is None) ^ (self.scaled_channels is None):
            raise ValueError(
                "scaled_num_blocks and scaled_channels must both be provided together"
            )

        if self.scaled_num_blocks is not None:
            self.scaled_num_blocks = _normalize_positive_int_sequence(
                self.scaled_num_blocks,
                "scaled_num_blocks",
                expected_length=2,
            )
            self.scaled_channels = _normalize_positive_int_sequence(
                self.scaled_channels,
                "scaled_channels",
                expected_length=2,
            )


def conv_3x3_bn(
    in_channels: int, out_channels: int, downsample: bool = False, **kwargs
) -> nn.Sequential:
    stride = 1 if not downsample else 2
    return nn.Sequential(
        nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        ),
        nn.BatchNorm2d(out_channels),
        nn.GELU(),
    )


class _PreNorm(nn.Module):
    def __init__(self, dim: int, fn: nn.Module, norm: Type[nn.Module]) -> Tensor:
        super().__init__()
        self.norm = norm(dim)
        self.fn = fn

    def forward(self, x: Tensor, **kwargs) -> Tensor:
        return self.fn(self.norm(x), **kwargs)


class _SEBlock(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, reduction: float = 0.25
    ) -> None:
        super().__init__()
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Sequential(
            nn.Linear(out_channels, int(in_channels * reduction), bias=False),
            nn.GELU(),
            nn.Linear(int(in_channels * reduction), out_channels, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        y = self.avgpool(x)
        y = y.reshape(y.shape[0], -1)
        y = self.fc(y).unsqueeze(axis=(-1, -2))

        return x * y


class _FeedForward(nn.Module):
    def __init__(self, dim: int, hidden_dim: int, dropout: float = 0.0) -> None:
        super().__init__()
        self.ff = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.ff(x)


class _MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        downsample: bool = False,
        expansion: int = 4,
        **kwargs,
    ) -> None:
        super().__init__()
        self.downsample = downsample
        stride = 1 if not self.downsample else 2
        hidden_dim = int(in_channels * expansion)

        if self.downsample:
            self.pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
            self.proj = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)

        if expansion == 1:
            conv = nn.Sequential(
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    kernel_size=3,
                    stride=stride,
                    padding=1,
                    groups=hidden_dim,
                    bias=False,
                ),
                nn.BatchNorm2d(hidden_dim),
                nn.GELU(),
                nn.Conv2d(hidden_dim, out_channels, kernel_size=1, bias=False),
                nn.BatchNorm2d(out_channels),
            )
        else:
            conv = nn.Sequential(
                nn.Conv2d(
                    in_channels, hidden_dim, kernel_size=1, stride=stride, bias=False
                ),
                nn.BatchNorm2d(hidden_dim),
                nn.GELU(),
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    kernel_size=3,
                    padding=1,
                    groups=hidden_dim,
                    bias=False,
                ),
                nn.BatchNorm2d(hidden_dim),
                nn.GELU(),
                _SEBlock(in_channels, hidden_dim),
                nn.Conv2d(hidden_dim, out_channels, kernel_size=1, bias=False),
                nn.BatchNorm2d(out_channels),
            )

        self.conv = _PreNorm(in_channels, conv, nn.BatchNorm2d)

    def forward(self, x: Tensor) -> Tensor:
        if self.downsample:
            return self.proj(self.pool(x)) + self.conv(x)
        else:
            return x + self.conv(x)


class _Attention(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        img_size: tuple[int, int],
        num_heads: int = 8,
        dim_head: int = 32,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        inner_dim = dim_head * num_heads
        project_out = not (num_heads == 1 and dim_head == in_channels)

        self.ih, self.iw = img_size
        self.num_heads = num_heads
        self.scale = dim_head**0.5

        self.relative_bias_table = nn.Parameter(
            lucid.zeros((2 * self.ih - 1) * (2 * self.iw - 1), num_heads)
        )

        coords = lucid.stack(
            lucid.meshgrid(lucid.arange(self.ih), lucid.arange(self.iw))
        )
        coords = coords.reshape(coords.shape[0], -1)
        relative_coords = coords[:, :, None] - coords[:, None, :]

        relative_coords[0] += self.ih - 1
        relative_coords[1] += self.iw - 1
        relative_coords[0] *= 2 * self.iw - 1

        relative_coords = lucid.einops.rearrange(relative_coords, "c h w -> h w c")
        relative_index = relative_coords.sum(axis=-1).flatten().unsqueeze(axis=1)

        self.relative_index: nn.Buffer
        self.register_buffer("relative_index", relative_index)

        self.attend = nn.Softmax(axis=-1)
        self.to_qkv = nn.Linear(in_channels, inner_dim * 3, bias=False)

        self.to_out = (
            nn.Sequential(nn.Linear(inner_dim, out_channels), nn.Dropout(dropout))
            if project_out
            else nn.Identity()
        )

    def forward(self, x: Tensor) -> Tensor:
        qkv = self.to_qkv(x).chunk(3, axis=-1)
        q, k, v = map(
            lambda tensor: lucid.einops.rearrange(
                tensor,
                "b n (h d) -> b h n d",
                h=self.num_heads,
                d=tensor.shape[-1] // self.num_heads,
            ),
            qkv,
        )
        dots = (q @ k.mT) * self.scale

        rel_idx = self.relative_index.repeat(self.num_heads, axis=1).to(x.device)
        col_indices = (
            (
                lucid.arange(self.relative_bias_table.shape[1])
                .unsqueeze(axis=0)
                .broadcast_to(rel_idx.shape)
            )
            .astype(lucid.Int)
            .to(x.device)
        )

        relative_bias = self.relative_bias_table[
            rel_idx.astype(lucid.Int), col_indices.astype(lucid.Int)
        ]
        relative_bias = lucid.einops.rearrange(
            relative_bias, "(h w) c -> c h w", h=self.ih * self.iw, w=self.ih * self.iw
        ).unsqueeze(axis=0)

        dots += relative_bias

        attn = self.attend(dots)
        out = attn @ v
        out = lucid.einops.rearrange(out, "b h n d -> b n (h d)")
        out = self.to_out(out)

        return out


class _Transformer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        img_size: tuple[int, int],
        num_heads: int = 8,
        dim_head: int = 32,
        downsample: bool = False,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        hidden_dim = int(in_channels * 4)

        self.ih, self.iw = img_size
        self.downsample = downsample

        if self.downsample:
            self.pool1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
            self.pool2 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
            self.proj = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)

        attn = _Attention(
            in_channels, out_channels, img_size, num_heads, dim_head, dropout
        )
        ff = _FeedForward(out_channels, hidden_dim, dropout)

        self.attn = nn.Sequential(
            nn.Rearrange("b c ih iw -> b (ih iw) c"),
            _PreNorm(in_channels, attn, nn.LayerNorm),
            nn.Rearrange("b (ih iw) c -> b c ih iw", ih=self.ih, iw=self.iw),
        )
        self.ff = nn.Sequential(
            nn.Rearrange("b c ih iw -> b (ih iw) c"),
            _PreNorm(out_channels, ff, nn.LayerNorm),
            nn.Rearrange("b (ih iw) c -> b c ih iw", ih=self.ih, iw=self.iw),
        )

    def forward(self, x: Tensor) -> Tensor:
        if self.downsample:
            x = self.proj(self.pool1(x)) + self.attn(self.pool2(x))
        else:
            x += self.attn(x)
        x += self.ff(x)
        return x


class CoAtNet(nn.Module):
    def __init__(self, config: CoAtNetConfig) -> None:
        super().__init__()
        self.config = config
        block = {"C": _MBConv, "T": _Transformer}
        get_block = lambda i: block[config.block_types[i]]

        self.ih, self.iw = config.img_size
        self.channels = config.channels
        self.num_heads = config.num_heads

        self._do_scale = (
            config.scaled_num_blocks is not None and config.scaled_channels is not None
        )

        self.s0 = self._make_layer(
            0, conv_3x3_bn, config.in_channels, config.channels[0], config.num_blocks[0]
        )
        self.s1 = self._make_layer(
            1,
            get_block(0),
            config.channels[0],
            config.channels[1],
            config.num_blocks[1],
        )
        self.s2 = self._make_layer(
            2,
            get_block(1),
            config.channels[1],
            config.channels[2],
            config.num_blocks[2],
        )
        self.s3 = (
            self._make_layer(
                3,
                get_block(2),
                config.channels[2],
                config.channels[3],
                config.num_blocks[3],
            )
            if not self._do_scale
            else nn.Identity()
        )
        self.s4 = self._make_layer(
            4,
            get_block(3),
            config.channels[3] if not self._do_scale else config.scaled_channels[1],
            config.channels[4],
            config.num_blocks[4],
        )

        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(config.channels[-1], config.num_classes, bias=False)

        if self._do_scale:
            self._scale_s3(config.scaled_num_blocks, config.scaled_channels)

    def _make_layer(
        self,
        i: int,
        block: Type[nn.Module],
        in_channels: int,
        out_channels: int,
        depth: int,
        downsample_first: bool = True,
    ) -> nn.Sequential:
        img_size = (self.ih // (2 ** (i + 1)), self.iw // (2 ** (i + 1)))

        layers = nn.ModuleList()
        for block_index in range(depth):
            new_block = block(
                in_channels if block_index == 0 else out_channels,
                out_channels,
                img_size=img_size,
                num_heads=self.num_heads,
                downsample=downsample_first if block_index == 0 else False,
            )
            layers.append(new_block)

        return nn.Sequential(*layers)

    def _scale_s3(
        self,
        depths: tuple[int, int],
        channels: tuple[int, int],
    ) -> None:
        self.s3_tandem = nn.Sequential(
            self._make_layer(3, _MBConv, self.channels[2], channels[0], depths[0]),
            self._make_layer(
                4,
                _Transformer,
                channels[0],
                channels[1],
                depths[1],
            ),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.s0(x)
        x = self.s1(x)
        x = self.s2(x)
        x = self.s3(x) if not self._do_scale else self.s3_tandem(x)
        x = self.s4(x)

        x = self.pool(x).reshape(-1, x.shape[1])
        x = self.fc(x)

        return x


def _raise_for_locked_factory_kwargs(
    kwargs: dict[str, object],
    locked_fields: set[str],
    message: str,
) -> None:
    if locked_fields & kwargs.keys():
        raise TypeError(message)


def _build_coatnet_config(
    *,
    img_size: tuple[int, int],
    in_channels: int,
    num_blocks: tuple[int, ...],
    channels: tuple[int, ...],
    num_classes: int,
    **kwargs,
) -> CoAtNetConfig:
    return CoAtNetConfig(
        img_size=img_size,
        in_channels=in_channels,
        num_blocks=num_blocks,
        channels=channels,
        num_classes=num_classes,
        **kwargs,
    )


@register_model
def coatnet_0(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"img_size", "in_channels", "num_blocks", "channels"},
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, or channels",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 3, 5, 2),
        channels=(64, 96, 192, 384, 768),
        num_classes=num_classes,
        **kwargs,
    )
    return CoAtNet(config)


@register_model
def coatnet_1(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"img_size", "in_channels", "num_blocks", "channels"},
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, or channels",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 6, 14, 2),
        channels=(64, 96, 192, 384, 768),
        num_classes=num_classes,
        **kwargs,
    )
    return CoAtNet(config)


@register_model
def coatnet_2(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"img_size", "in_channels", "num_blocks", "channels"},
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, or channels",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 6, 14, 2),
        channels=(128, 128, 256, 512, 1024),
        num_classes=num_classes,
        **kwargs,
    )
    return CoAtNet(config)


@register_model
def coatnet_3(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"img_size", "in_channels", "num_blocks", "channels"},
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, or channels",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 6, 14, 2),
        channels=(192, 192, 384, 768, 1536),
        num_classes=num_classes,
        **kwargs,
    )
    return CoAtNet(config)


@register_model
def coatnet_4(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"img_size", "in_channels", "num_blocks", "channels"},
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, or channels",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 12, 28, 2),
        channels=(192, 192, 384, 768, 1536),
        num_classes=num_classes,
        **kwargs,
    )
    return CoAtNet(config)


@register_model
def coatnet_5(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"img_size", "in_channels", "num_blocks", "channels", "num_heads"},
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, channels, or num_heads",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 12, 28, 2),
        channels=(192, 256, 512, 1280, 2048),
        num_classes=num_classes,
        num_heads=64,
        **kwargs,
    )
    return CoAtNet(config)


@register_model
def coatnet_6(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {
            "img_size",
            "in_channels",
            "num_blocks",
            "channels",
            "num_heads",
            "scaled_num_blocks",
            "scaled_channels",
        },
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, channels, num_heads, scaled_num_blocks, or scaled_channels",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 4, 8, 2),
        channels=(192, 192, 384, 768, 2048),
        num_classes=num_classes,
        num_heads=128,
        scaled_num_blocks=(8, 42),
        scaled_channels=(768, 1536),
        **kwargs,
    )
    return CoAtNet(config)


@register_model
def coatnet_7(num_classes: int = 1000, **kwargs) -> CoAtNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {
            "img_size",
            "in_channels",
            "num_blocks",
            "channels",
            "num_heads",
            "scaled_num_blocks",
            "scaled_channels",
        },
        "factory variants do not allow overriding preset img_size, in_channels, num_blocks, channels, num_heads, scaled_num_blocks, or scaled_channels",
    )
    config = _build_coatnet_config(
        img_size=(224, 224),
        in_channels=3,
        num_blocks=(2, 2, 4, 8, 2),
        channels=(192, 256, 512, 1024, 3072),
        num_classes=num_classes,
        num_heads=128,
        scaled_num_blocks=(8, 42),
        scaled_channels=(1024, 2048),
        **kwargs,
    )
    return CoAtNet(config)



from typing import Literal, Callable

import lucid
from lucid._tensor import Tensor


def affine_grid(
    theta: Tensor, size: tuple[int, ...], align_corners: bool = True
) -> Tensor:
    N, _, H, W = size
    device = theta.device

    if align_corners:
        xs = lucid.linspace(-1, 1, W)
        ys = lucid.linspace(-1, 1, H)
    else:
        xs = lucid.linspace(-1 + 1 / W, 1 - 1 / W, W)
        ys = lucid.linspace(-1 + 1 / H, 1 - 1 / H, H)

    x, y = lucid.meshgrid(xs, ys)
    ones = lucid.ones_like(x)

    grid = lucid.stack([x, y, ones], axis=-1)
    grid = grid.reshape(1, H * W, 3).repeat(N, axis=0)
    grid = grid.astype(lucid.Float).to(device).free()

    theta = theta.reshape(N, 2, 3)
    out = grid @ theta.transpose((0, 2, 1))
    out = out.reshape(N, H, W, 2)

    return out


_PaddingType = Literal["zeros", "border"]
_InterpolateType = Literal["bilinear", "nearest"]


def grid_sample(
    input_: Tensor,
    grid: Tensor,
    mode: _InterpolateType = "bilinear",
    padding_mode: _PaddingType = "zeros",
    align_corners: bool = True,
) -> Tensor:
    N, C, H_in, W_in = input_.shape
    N_grid, H_out, W_out, _ = grid.shape
    assert N == N_grid, "Batch size mismatch"

    if align_corners:
        ix = (grid[..., 0] + 1) * (W_in - 1) / 2
        iy = (grid[..., 1] + 1) * (H_in - 1) / 2
    else:
        ix = (grid[..., 0] + 1) * W_in / 2 - 0.5
        iy = (grid[..., 1] + 1) * H_in / 2 - 0.5

    if mode == "nearest":
        ix = ix.round()
        iy = iy.round()

        valid: Tensor | None = None
        if padding_mode == "border":
            ix = lucid.clip(ix, 0, W_in - 1)
            iy = lucid.clip(iy, 0, H_in - 1)
        elif padding_mode == "zeros":
            valid = (
                (ix >= 0) & (ix <= W_in - 1) & (iy >= 0) & (iy <= H_in - 1)
            ).astype(input_.dtype)
            ix = lucid.clip(ix, 0, W_in - 1)
            iy = lucid.clip(iy, 0, H_in - 1)
        else:
            raise ValueError(f"Unsupported padding_mode: {padding_mode}")

        ix_idx = ix.astype(lucid.Int)
        iy_idx = iy.astype(lucid.Int)

        n_idx = (
            lucid.arange(N, device=input_.device).reshape(N, 1, 1, 1).astype(lucid.Int)
        )
        c_idx = (
            lucid.arange(C, device=input_.device).reshape(1, C, 1, 1).astype(lucid.Int)
        )

        n_idx = n_idx.repeat(C, axis=1).repeat(H_out, axis=2).repeat(W_out, axis=3)
        c_idx = c_idx.repeat(N, axis=0).repeat(H_out, axis=2).repeat(W_out, axis=3)
        iy_idx = iy_idx[:, None, :, :].repeat(C, axis=1)
        ix_idx = ix_idx[:, None, :, :].repeat(C, axis=1)

        output = input_[n_idx, c_idx, iy_idx, ix_idx]
        if valid is not None:
            output = output * valid[:, None, :, :]
        return output

    elif mode == "bilinear":
        if padding_mode == "border":
            ix = lucid.clip(ix, 0, W_in - 1)
            iy = lucid.clip(iy, 0, H_in - 1)
        elif padding_mode != "zeros":
            raise ValueError(f"Unsupported padding_mode: {padding_mode}")

        x0 = ix.floor()
        x1 = x0 + 1
        y0 = iy.floor()
        y1 = y0 + 1

        x0_idx = lucid.clip(x0, 0, W_in - 1).astype(lucid.Int)
        x1_idx = lucid.clip(x1, 0, W_in - 1).astype(lucid.Int)
        y0_idx = lucid.clip(y0, 0, H_in - 1).astype(lucid.Int)
        y1_idx = lucid.clip(y1, 0, H_in - 1).astype(lucid.Int)

        wa = (x1 - ix) * (y1 - iy)
        wb = (x1 - ix) * (iy - y0)
        wc = (ix - x0) * (y1 - iy)
        wd = (ix - x0) * (iy - y0)

        n_idx = (
            lucid.arange(N, device=input_.device).reshape(N, 1, 1, 1).astype(lucid.Int)
        )
        c_idx = (
            lucid.arange(C, device=input_.device).reshape(1, C, 1, 1).astype(lucid.Int)
        )

        n_idx = n_idx.repeat(C, axis=1).repeat(H_out, axis=2).repeat(W_out, axis=3)
        c_idx = c_idx.repeat(N, axis=0).repeat(H_out, axis=2).repeat(W_out, axis=3)

        def _gather(y: Tensor, x: Tensor) -> Tensor:
            y = y[:, None, :, :].repeat(C, axis=1)
            x = x[:, None, :, :].repeat(C, axis=1)

            return input_[n_idx, c_idx, y, x]

        Ia = _gather(y0_idx, x0_idx)
        Ib = _gather(y1_idx, x0_idx)
        Ic = _gather(y0_idx, x1_idx)
        Id = _gather(y1_idx, x1_idx)

        if padding_mode == "zeros":
            mask_a = (
                (x0 >= 0) & (x0 <= W_in - 1) & (y0 >= 0) & (y0 <= H_in - 1)
            ).astype(input_.dtype)
            mask_b = (
                (x0 >= 0) & (x0 <= W_in - 1) & (y1 >= 0) & (y1 <= H_in - 1)
            ).astype(input_.dtype)
            mask_c = (
                (x1 >= 0) & (x1 <= W_in - 1) & (y0 >= 0) & (y0 <= H_in - 1)
            ).astype(input_.dtype)
            mask_d = (
                (x1 >= 0) & (x1 <= W_in - 1) & (y1 >= 0) & (y1 <= H_in - 1)
            ).astype(input_.dtype)

            Ia = Ia * mask_a[:, None, :, :]
            Ib = Ib * mask_b[:, None, :, :]
            Ic = Ic * mask_c[:, None, :, :]
            Id = Id * mask_d[:, None, :, :]

        wa = wa[:, None, :, :].repeat(C, axis=1)
        wb = wb[:, None, :, :].repeat(C, axis=1)
        wc = wc[:, None, :, :].repeat(C, axis=1)
        wd = wd[:, None, :, :].repeat(C, axis=1)

        output = Ia * wa + Ib * wb + Ic * wc + Id * wd
        return output

    else:
        raise ValueError(f"Unsupported interpolation mode: {mode}")



import unicodedata


def is_whitespace(ch: str) -> bool:
    if ch in (" ", "\t", "\n", "\r"):
        return True
    return unicodedata.category(ch) == "Zs"


def is_control(ch: str) -> bool:
    if ch in ("\t", "\n", "\r"):
        return False
    return unicodedata.category(ch) in {"Cc", "Cf"}


def is_punctuation(ch: str) -> bool:
    cp = ord(ch)
    if (33 <= cp <= 47) or (58 <= cp <= 64) or (91 <= cp <= 96) or (123 <= cp <= 126):
        return True
    return unicodedata.category(ch).startswith("P")


def clean_text(text: str) -> str:
    out: list[str] = []
    for ch in text:
        cp = ord(ch)
        if cp in (0, 0xFFFD) or is_control(ch):
            continue
        if is_whitespace(ch):
            out.append(" ")
        else:
            out.append(ch)
    return "".join(out)


def split_on_punctuation(token: str) -> list[str]:
    out: list[str] = []
    current: list[str] = []
    for ch in token:
        if is_punctuation(ch):
            if current:
                out.append("".join(current))
                current = []
            out.append(ch)
        else:
            current.append(ch)
    if current:
        out.append("".join(current))
    return out


def basic_tokenize(text: str, lowercase: bool = True) -> list[str]:
    tokens: list[str] = []
    for token in text.strip().split():
        if lowercase:
            token = token.lower()
        tokens.extend(split_on_punctuation(token))
    return tokens



from typing import Any, Callable, ClassVar

import numpy as np
import torch

import lucid
from lucid._func import bfunc
from lucid.test.core import TensorOpTorchCase, TensorOpWithTorchBase

_ARITH_LEFT: list[list[float]] = [[1.0, 2.0], [3.0, 4.0]]
_ARITH_RIGHT: list[list[float]] = [[2.0, 1.0], [0.0, -1.0]]

_CMP_LEFT: list[list[float]] = _ARITH_LEFT
_CMP_RIGHT: list[list[float]] = [[2.0, 2.0], [3.0, 1.0]]

_MATRIX_RIGHT: list[list[float]] = [[5.0, 6.0], [7.0, 8.0]]

_BITWISE_LEFT: list[list[int]] = [[1, 2], [3, 4]]
_BITWISE_RIGHT: list[list[int]] = [[2, 1], [0, 7]]


def _bind_bfunc(op: type[Any]) -> Callable[[Any, Any], Any]:
    def _forward(left: Any, right: Any) -> Any:
        return op()(left, right)

    return _forward


class _TestBinaryFuncOpBase(TensorOpWithTorchBase):
    _collect_nested_test_classes = True
    case_name: ClassVar[str]
    left: ClassVar[Any]
    right: ClassVar[Any]

    forward_op: ClassVar[Callable[[Any, Any], Any]]
    torch_forward: ClassVar[Callable[[Any, Any], Any]]
    expected_forward: ClassVar[Any]

    requires_grad: ClassVar[bool | tuple[bool, bool]] = True
    dtype: ClassVar[Any | None] = None
    compare_grad_with_torch: ClassVar[bool] = True

    check_finite: ClassVar[bool] = True
    expected_input_grads: ClassVar[tuple[object | None, ...] | None] = None

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu", "gpu")

    def _build_binary_inputs(
        self,
        left: Any,
        right: Any,
        *,
        dtype: Any | None = None,
        requires_grad: bool | tuple[bool, bool] = True,
    ) -> tuple[Any, ...]:
        if isinstance(requires_grad, bool):
            requires_grad = (requires_grad, requires_grad)

        if dtype is None:
            return (
                self.tensor(left, requires_grad=requires_grad[0]),
                self.tensor(right, requires_grad=requires_grad[1]),
            )

        return (
            self.tensor(left, requires_grad=requires_grad[0], dtype=dtype),
            self.tensor(right, requires_grad=requires_grad[1], dtype=dtype),
        )

    def _case_grad_requires(self) -> tuple[bool, bool]:
        if isinstance(self.requires_grad, tuple):
            return self.requires_grad
        return (self.requires_grad, self.requires_grad)

    def _build_case(self) -> TensorOpTorchCase:
        forward_op = vars(type(self)).get("forward_op")
        torch_forward = vars(type(self)).get("torch_forward")

        if isinstance(forward_op, staticmethod):
            forward_op = forward_op.__func__
        if isinstance(torch_forward, staticmethod):
            torch_forward = torch_forward.__func__

        return TensorOpTorchCase(
            name=self.case_name,
            build_inputs=lambda _case: self._build_binary_inputs(
                self.left,
                self.right,
                dtype=self.dtype,
                requires_grad=self._case_grad_requires(),
            ),
            forward_op=forward_op,
            expected_forward=self.expected_forward,
            expected_input_grads=self.expected_input_grads,
            requires_grad=self._case_grad_requires(),
            check_finite=self.check_finite,
            torch_forward=torch_forward,
            compare_grad_with_torch=self.compare_grad_with_torch,
        )

    def tensor_op_cases(self) -> tuple[TensorOpTorchCase, ...]:
        return (self._build_case(),)


class _TestBinaryNoGradFuncOpBase(_TestBinaryFuncOpBase):
    requires_grad: ClassVar[tuple[bool, bool]] = (False, False)
    compare_grad_with_torch: ClassVar[bool] = False


class TestAdd(_TestBinaryFuncOpBase):
    case_name = "add"
    left = _ARITH_LEFT
    right = _ARITH_RIGHT
    forward_op = lucid.add
    torch_forward = torch.add
    expected_forward = np.array([[3.0, 3.0], [3.0, 3.0]], dtype=np.float64)


class TestSub(_TestBinaryFuncOpBase):
    case_name = "sub"
    left = _ARITH_LEFT
    right = _ARITH_RIGHT
    forward_op = lucid.sub
    torch_forward = torch.sub
    expected_forward = np.array([[-1.0, 1.0], [3.0, 5.0]], dtype=np.float64)


class TestMultiply(_TestBinaryFuncOpBase):
    case_name = "mul"
    left = _ARITH_LEFT
    right = _ARITH_RIGHT
    forward_op = lucid.multiply
    torch_forward = torch.mul
    expected_forward = np.array([[2.0, 2.0], [0.0, -4.0]], dtype=np.float64)


class TestDiv(_TestBinaryFuncOpBase):
    __test__ = False

    class TestTruediv(_TestBinaryFuncOpBase):
        case_name = "truediv"
        left = [[1.0, 2.0], [3.0, 3.0]]
        right = [[2.0, 1.0], [2.0, 2.0]]
        forward_op = lucid.div
        torch_forward = torch.div
        expected_forward = np.array([[0.5, 2.0], [1.5, 1.5]], dtype=np.float64)

    class TestFloorDiv(_TestBinaryNoGradFuncOpBase):
        case_name = "floordiv_no_grad"
        left = [[1, 2], [3, 5]]
        right = [[2, 3], [2, 2]]
        forward_op = staticmethod(lambda a, b: lucid.div(a, b, floor=True))
        torch_forward = torch.floor_divide
        dtype = lucid.Int64
        expected_forward = np.array([[0, 0], [1, 2]], dtype=np.int64)


class TestMinimum(_TestBinaryFuncOpBase):
    case_name = "minimum"
    left = _ARITH_LEFT
    right = _ARITH_RIGHT
    forward_op = lucid.minimum
    torch_forward = torch.minimum
    expected_forward = np.array([[1.0, 1.0], [0.0, -1.0]], dtype=np.float64)


class TestMaximum(_TestBinaryFuncOpBase):
    case_name = "maximum"
    left = _ARITH_LEFT
    right = _ARITH_RIGHT
    forward_op = lucid.maximum
    torch_forward = torch.maximum
    expected_forward = np.array([[2.0, 2.0], [3.0, 4.0]], dtype=np.float64)


class TestPower(_TestBinaryFuncOpBase):
    case_name = "power"
    left = [[1.0, 2.0], [3.0, 4.0]]
    right = [[2.0, 3.0], [1.0, 2.0]]
    forward_op = lucid.power
    torch_forward = torch.pow
    expected_forward = np.array([[1.0, 8.0], [3.0, 16.0]], dtype=np.float64)


class TestDot(_TestBinaryFuncOpBase):
    case_name = "dot"
    left = [1.0, 2.0, 3.0]
    right = [4.0, 5.0, 6.0]
    forward_op = lucid.dot
    torch_forward = torch.dot
    expected_forward = np.array(32.0)


class TestInner(_TestBinaryFuncOpBase):
    case_name = "inner"
    left = _ARITH_LEFT
    right = _MATRIX_RIGHT
    forward_op = lucid.inner
    torch_forward = torch.inner
    expected_forward = np.array([[17.0, 23.0], [39.0, 53.0]], dtype=np.float64)


class TestOuter(_TestBinaryFuncOpBase):
    case_name = "outer"
    left = [1.0, 2.0, 3.0]
    right = [4.0, 5.0]
    forward_op = lucid.outer
    torch_forward = torch.outer
    expected_forward = np.array(
        [[4.0, 5.0], [8.0, 10.0], [12.0, 15.0]], dtype=np.float64
    )


class TestMatmul(_TestBinaryFuncOpBase):
    case_name = "matmul"
    left = _ARITH_LEFT
    right = [[1.0, 0.0], [2.0, 1.0]]
    forward_op = lucid.matmul
    torch_forward = torch.matmul
    expected_forward = np.array([[5.0, 2.0], [11.0, 4.0]], dtype=np.float64)


class TestTensorDot(_TestBinaryFuncOpBase):
    case_name = "tensordot"
    left = _ARITH_LEFT
    right = _MATRIX_RIGHT
    forward_op = lucid.tensordot
    torch_forward = lambda a, b: torch.tensordot(a, b, dims=2)
    expected_forward = np.array(70.0)


class TestEqual(_TestBinaryNoGradFuncOpBase):
    case_name = "equal_no_grad"
    left = _CMP_LEFT
    right = _CMP_RIGHT
    forward_op = _bind_bfunc(bfunc._equal)
    torch_forward = torch.eq
    expected_forward = np.array([[False, True], [True, False]], dtype=bool)


class TestNotEqual(_TestBinaryNoGradFuncOpBase):
    case_name = "not_equal_no_grad"
    left = _CMP_LEFT
    right = _CMP_RIGHT
    forward_op = _bind_bfunc(bfunc._not_equal)
    torch_forward = torch.ne
    expected_forward = np.array([[True, False], [False, True]], dtype=bool)


class TestGreater(_TestBinaryNoGradFuncOpBase):
    case_name = "greater_no_grad"
    left = _CMP_LEFT
    right = _CMP_RIGHT
    forward_op = _bind_bfunc(bfunc._greater)
    torch_forward = torch.gt
    expected_forward = np.array([[False, False], [False, True]], dtype=bool)


class TestGreaterOrEqual(_TestBinaryNoGradFuncOpBase):
    case_name = "greater_equal_no_grad"
    left = _CMP_LEFT
    right = _CMP_RIGHT
    forward_op = _bind_bfunc(bfunc._greater_or_equal)
    torch_forward = torch.ge
    expected_forward = np.array([[False, True], [True, True]], dtype=bool)


class TestLess(_TestBinaryNoGradFuncOpBase):
    case_name = "less_no_grad"
    left = _CMP_LEFT
    right = _CMP_RIGHT
    forward_op = _bind_bfunc(bfunc._less)
    torch_forward = torch.lt
    expected_forward = np.array([[True, False], [False, False]], dtype=bool)


class TestLessOrEqual(_TestBinaryNoGradFuncOpBase):
    case_name = "less_equal_no_grad"
    left = _CMP_LEFT
    right = _CMP_RIGHT
    forward_op = _bind_bfunc(bfunc._less_or_equal)
    torch_forward = torch.le
    expected_forward = np.array([[True, True], [True, False]], dtype=bool)


class TestBitwiseAnd(_TestBinaryNoGradFuncOpBase):
    case_name = "bitwise_and_no_grad"
    left = _BITWISE_LEFT
    right = _BITWISE_RIGHT
    forward_op = _bind_bfunc(bfunc._bitwise_and)
    torch_forward = torch.bitwise_and
    dtype = lucid.Int64
    expected_forward = np.array([[0, 0], [0, 4]], dtype=np.int64)


class TestBitwiseOr(_TestBinaryNoGradFuncOpBase):
    case_name = "bitwise_or_no_grad"
    left = _BITWISE_LEFT
    right = _BITWISE_RIGHT
    forward_op = _bind_bfunc(bfunc._bitwise_or)
    torch_forward = torch.bitwise_or
    dtype = lucid.Int64
    expected_forward = np.array([[3, 3], [3, 7]], dtype=np.int64)



from dataclasses import dataclass, field
from typing import Literal, Sequence

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.vision import unet as unet

__all__ = [
    "AttentionUNetGateConfig",
    "AttentionUNetConfig",
    "AttentionUNet2d",
    "AttentionUNet3d",
]


_AttentionGateMode = Literal["additive"]
_AttentionGateAct = Literal["relu"]
_AttentionCoeffAct = Literal["sigmoid"]
_AttentionResample = Literal["bilinear", "trilinear", "nearest"]


@dataclass
class AttentionUNetGateConfig:
    enabled: bool = True

    mode: _AttentionGateMode = "additive"
    gate_activation: _AttentionGateAct = "relu"
    attention_activation: _AttentionCoeffAct = "sigmoid"
    use_grid_attention: bool = True

    inter_channels: int | tuple[int, ...] | list[int] | None = None
    attention_channels: int = 1

    resample_mode: _AttentionResample = "bilinear"
    gate_on_skips: tuple[bool, ...] | list[bool] | None = None

    skip_low_level_gates: bool = False
    use_multi_scale_gating: bool = True

    project_skip_with_1x1: bool = True
    project_gating_with_1x1: bool = True

    init_pass_through: bool = True

    def __post_init__(self) -> None:
        if self.attention_channels <= 0:
            raise ValueError("attention_channels must be greater than 0.")

        if isinstance(self.inter_channels, list):
            self.inter_channels = tuple(self.inter_channels)
        if isinstance(self.inter_channels, tuple):
            if len(self.inter_channels) == 0:
                raise ValueError("inter_channels must not be empty.")
            if any(ch <= 0 for ch in self.inter_channels):
                raise ValueError("inter_channels values must be positive.")

        if self.gate_on_skips is not None:
            self.gate_on_skips = tuple(self.gate_on_skips)
            if len(self.gate_on_skips) == 0:
                raise ValueError("gate_on_skips must not be empty.")


@dataclass
class AttentionUNetConfig(unet.UNetConfig):
    block: Literal["basic"] = "basic"
    norm: unet._UNetNormName = "batch"
    act: unet._UNetActName = "relu"

    skip_merge: Literal["concat"] = "concat"
    downsample_mode: unet._UNetDownsample = "maxpool"
    upsample_mode: unet._UNetUpsample = "bilinear"

    deep_supervision: bool = True
    attention: AttentionUNetGateConfig = field(default_factory=AttentionUNetGateConfig)

    def __post_init__(self) -> None:
        super().__post_init__()
        if not isinstance(self.attention, AttentionUNetGateConfig):
            raise TypeError("attention must be an AttentionUNetGateConfig.")

        decoder_depth = len(self.decoder_stages)

        if self.attention.gate_on_skips is None:
            if decoder_depth == 0:
                self.attention.gate_on_skips = tuple()
            elif self.attention.skip_low_level_gates:
                self.attention.gate_on_skips = (True,) * decoder_depth
            else:
                self.attention.gate_on_skips = (True,) * max(decoder_depth - 1, 0) + (
                    False,
                )
        else:
            if len(self.attention.gate_on_skips) != decoder_depth:
                raise ValueError(
                    "attention.gate_on_skips must have 'len(decoder_stages)' entries."
                )

        if isinstance(self.attention.inter_channels, tuple):
            num_gated = sum(self.attention.gate_on_skips)
            if len(self.attention.inter_channels) not in (decoder_depth, num_gated):
                raise ValueError(
                    "attention.inter_channels must match decoder depth or "
                    "the number of enabled gates."
                )

        if self.block != "basic":
            raise ValueError("Attention U-Net requires block='basic'")
        if self.skip_merge != "concat":
            raise ValueError("Attention U-Net requires skip_merge='concat'")
        if self.attention.mode != "additive":
            raise ValueError("Attention U-Net requires additive gates")
        if self.attention.attention_activation != "sigmoid":
            raise ValueError("Attention U-Net requires sigmoid gate coefficients")
        if not self.attention.use_grid_attention:
            raise ValueError("Attention U-Net requires grid attention")


class _IdentityAttentionGate(nn.Module):
    def forward(self, skip: Tensor, *args, **kwargs) -> Tensor:
        return skip


class _AttentionGate2d(nn.Module):
    def __init__(
        self,
        skip_channels: int,
        gating_channels: int,
        inter_channels: int,
        gate_config: AttentionUNetGateConfig,
        *,
        align_corners: bool,
    ) -> None:
        super().__init__()
        self.gate_config = gate_config
        self.align_corners = align_corners

        gate_bias = True
        self.theta_x = (
            nn.Conv2d(skip_channels, inter_channels, kernel_size=1, bias=gate_bias)
            if gate_config.project_skip_with_1x1
            else nn.Identity()
        )
        self.phi_g = (
            nn.Conv2d(gating_channels, inter_channels, kernel_size=1, bias=gate_bias)
            if gate_config.project_gating_with_1x1
            else nn.Identity()
        )
        self.psi = nn.Conv2d(
            inter_channels, gate_config.attention_channels, kernel_size=1, bias=True
        )

        if gate_config.gate_activation == "relu":
            self.activation = nn.ReLU()
        else:
            raise ValueError(
                f"Unsupported gate activation: '{gate_config.gate_activation}'"
            )

        if gate_config.attention_activation == "sigmoid":
            self.coeff_activation = nn.Sigmoid()
        else:
            raise ValueError("Attention U-Net requries sigmoid attention coefficients.")

        self.alpha_proj = (
            nn.Conv2d(
                gate_config.attention_channels, skip_channels, kernel_size=1, bias=True
            )
            if gate_config.attention_channels != 1
            else None
        )

        if gate_config.init_pass_through:
            nn.init.constant(self.psi.weight, 0.0)
            if self.psi.bias is not None:
                nn.init.constant(self.psi.bias, 2.0)

            if self.alpha_proj is not None:
                nn.init.constant(self.alpha_proj.weight, 0.0)
                if self.alpha_proj.bias is not None:
                    nn.init.constant(self.alpha_proj.bias, 1.0)

    def _resize(
        self, x: Tensor, size: tuple[int, int], mode: _AttentionResample | None = None
    ) -> Tensor:
        mode = self.gate_config.resample_mode if mode is None else mode
        return F.interpolate(
            x,
            size=size,
            mode=mode,
            align_corners=self.align_corners if mode == "bilinear" else False,
        )

    def forward(self, skip: Tensor, gating: Tensor) -> Tensor:
        theta_x = self.theta_x(skip)
        phi_g = self.phi_g(gating)

        if theta_x.shape[1] != phi_g.shape[1]:
            raise ValueError(
                "Projected skip and gating features must have same channel width."
            )

        if self.gate_config.use_grid_attention:
            phi_g = self._resize(phi_g, theta_x.shape[-2:])
        else:
            phi_g = lucid.mean(phi_g, axis=(-2, -1), keepdims=True)
            phi_g = lucid.broadcast_to(phi_g, theta_x.shape)

        f = self.activation(theta_x + phi_g)
        alpha = self.coeff_activation(self.psi(f))

        if self.alpha_proj is not None:
            alpha = self.alpha_proj(alpha)

        alpha = self._resize(alpha, skip.shape[-2:])
        return alpha * skip


class _AttentionGate3d(nn.Module):
    def __init__(
        self,
        skip_channels: int,
        gating_channels: int,
        inter_channels: int,
        gate_config: AttentionUNetGateConfig,
        *,
        align_corners: bool,
    ) -> None:
        super().__init__()
        self.gate_config = gate_config
        self.align_corners = align_corners

        gate_bias = True
        self.theta_x = (
            nn.Conv3d(skip_channels, inter_channels, kernel_size=1, bias=gate_bias)
            if gate_config.project_skip_with_1x1
            else nn.Identity()
        )
        self.phi_g = (
            nn.Conv3d(gating_channels, inter_channels, kernel_size=1, bias=gate_bias)
            if gate_config.project_gating_with_1x1
            else nn.Identity()
        )
        self.psi = nn.Conv3d(
            inter_channels, gate_config.attention_channels, kernel_size=1, bias=True
        )

        if gate_config.gate_activation == "relu":
            self.activation = nn.ReLU()
        else:
            raise ValueError(
                f"Unsupported gate activation: '{gate_config.gate_activation}'"
            )

        if gate_config.attention_activation == "sigmoid":
            self.coeff_activation = nn.Sigmoid()
        else:
            raise ValueError("Attention U-Net requires sigmoid attention coefficients.")

        self.alpha_proj = (
            nn.Conv3d(
                gate_config.attention_channels, skip_channels, kernel_size=1, bias=True
            )
            if gate_config.attention_channels != 1
            else None
        )

        if gate_config.init_pass_through:
            nn.init.constant(self.psi.weight, 0.0)
            if self.psi.bias is not None:
                nn.init.constant(self.psi.bias, 2.0)

            if self.alpha_proj is not None:
                nn.init.constant(self.alpha_proj.weight, 0.0)
                if self.alpha_proj.bias is not None:
                    nn.init.constant(self.alpha_proj.bias, 1.0)

    def _resize(
        self,
        x: Tensor,
        size: tuple[int, int, int],
        mode: _AttentionResample | None = None,
    ) -> Tensor:
        mode = self.gate_config.resample_mode if mode is None else mode
        if mode == "bilinear":
            mode = "trilinear"
        return F.interpolate(
            x,
            size=size,
            mode=mode,
            align_corners=self.align_corners if mode == "trilinear" else False,
        )

    def forward(self, skip: Tensor, gating: Tensor) -> Tensor:
        theta_x = self.theta_x(skip)
        phi_g = self.phi_g(gating)

        if theta_x.shape[1] != phi_g.shape[1]:
            raise ValueError(
                "Projected skip and gating features must have same channel width."
            )

        if self.gate_config.use_grid_attention:
            phi_g = self._resize(phi_g, theta_x.shape[-3:])
        else:
            phi_g = lucid.mean(phi_g, axis=(-3, -2, -1), keepdims=True)
            phi_g = lucid.broadcast_to(phi_g, theta_x.shape)

        f = self.activation(theta_x + phi_g)
        alpha = self.coeff_activation(self.psi(f))

        if self.alpha_proj is not None:
            alpha = self.alpha_proj(alpha)

        alpha = self._resize(alpha, skip.shape[-3:])
        return alpha * skip


class AttentionUNet2d(unet.UNet2d):
    def __init__(self, config: AttentionUNetConfig) -> None:
        super().__init__(config)
        self.config = config

        if not isinstance(config, AttentionUNetConfig):
            raise TypeError("config must be an AttentionUNetConfig.")
        if not isinstance(config.attention, AttentionUNetGateConfig):
            raise TypeError("config.attention must be an AttentionUNetGateConfig.")

        self.attention_gates: Sequence[_IdentityAttentionGate | _AttentionGate2d] = (
            nn.ModuleList()
        )

        decoder_depth = len(config.decoder_stages)
        gate_on_skips = config.attention.gate_on_skips
        if gate_on_skips is None:
            raise ValueError("attention.gate_on_skips must be resolved in config.")

        skip_channels = tuple(
            stage.channels for stage in reversed(config.encoder_stages[:-1])
        )
        gating_channels = (
            config.bottleneck.channels,
            *(stage.channels for stage in config.decoder_stages[:-1]),
        )

        for idx in range(decoder_depth):
            enabled = config.attention.enabled and gate_on_skips[idx]
            if not enabled:
                self.attention_gates.append(_IdentityAttentionGate())
                continue

            self.attention_gates.append(
                _AttentionGate2d(
                    skip_channels=skip_channels[idx],
                    gating_channels=gating_channels[idx],
                    inter_channels=self._resolve_gate_inter_channels(
                        idx, skip_channels[idx], gating_channels[idx]
                    ),
                    gate_config=config.attention,
                    align_corners=config.align_corners,
                )
            )

    def _resolve_gate_inter_channels(
        self, idx: int, skip_channels: int, gating_channels: int
    ) -> int:
        inter_channels = self.config.attention.inter_channels
        if inter_channels is None:
            return max(1, min(skip_channels, gating_channels) // 2)

        if isinstance(inter_channels, int):
            return inter_channels

        enabled_indices = [
            i
            for i, enabled in enumerate(self.config.attention.gate_on_skips)
            if enabled
        ]

        if len(inter_channels) == len(self.config.decoder_stages):
            return inter_channels[idx]

        enabled_pos = enabled_indices.index(idx)
        return inter_channels[enabled_pos]

    def forward(self, x: Tensor) -> Tensor | dict[str, Tensor | list[Tensor]]:
        input_size = x.shape[-2:]
        x = self.stem(x)

        skips: list[Tensor] = []
        for idx, stage in enumerate(self.encoder):
            x = stage(x)
            skips.append(x)

            if idx < len(self.downsamplers):
                x = self.downsamplers[idx](x)

        x = self.bottleneck(x)

        aux_outputs: list[Tensor] = []
        decode_skips = list(reversed(skips[:-1]))

        for idx, (gate, upsample, stage, skip) in enumerate(
            zip(self.attention_gates, self.upsamplers, self.decoder, decode_skips)
        ):
            gated_skip = gate(skip, x)
            x = upsample(x, target_size=gated_skip.shape[-2:])
            x = self._merge_skip(x, gated_skip)
            x = stage(x)

            if self.config.deep_supervision and idx < len(self.aux_heads):
                aux = self.aux_heads[idx](x)
                aux_outputs.append(self._resize_to_input(aux, input_size))

        out = self.head(x)
        out = self._resize_to_input(out, input_size)

        if self.config.deep_supervision:
            return {"out": out, "aux": aux_outputs}

        return out


class AttentionUNet3d(unet.UNet3d):
    def __init__(self, config: AttentionUNetConfig) -> None:
        super().__init__(config)
        self.config = config

        if not isinstance(config, AttentionUNetConfig):
            raise TypeError("config must be an AttentionUNetConfig.")
        if not isinstance(config.attention, AttentionUNetGateConfig):
            raise TypeError("config.attention must be an AttentionUNetGateConfig.")

        self.attention_gates: Sequence[_IdentityAttentionGate | _AttentionGate3d] = (
            nn.ModuleList()
        )

        decoder_depth = len(config.decoder_stages)
        gate_on_skips = config.attention.gate_on_skips
        if gate_on_skips is None:
            raise ValueError("attention.gate_on_skips must be resolved in config.")

        skip_channels = tuple(
            stage.channels for stage in reversed(config.encoder_stages[:-1])
        )
        gating_channels = (
            config.bottleneck.channels,
            *(stage.channels for stage in config.decoder_stages[:-1]),
        )

        for idx in range(decoder_depth):
            enabled = config.attention.enabled and gate_on_skips[idx]
            if not enabled:
                self.attention_gates.append(_IdentityAttentionGate())
                continue

            self.attention_gates.append(
                _AttentionGate3d(
                    skip_channels=skip_channels[idx],
                    gating_channels=gating_channels[idx],
                    inter_channels=self._resolve_gate_inter_channels(
                        idx, skip_channels[idx], gating_channels[idx]
                    ),
                    gate_config=config.attention,
                    align_corners=config.align_corners,
                )
            )

    def _resolve_gate_inter_channels(
        self, idx: int, skip_channels: int, gating_channels: int
    ) -> int:
        inter_channels = self.config.attention.inter_channels
        if inter_channels is None:
            return max(1, min(skip_channels, gating_channels) // 2)

        if isinstance(inter_channels, int):
            return inter_channels

        enabled_indices = [
            i
            for i, enabled in enumerate(self.config.attention.gate_on_skips)
            if enabled
        ]

        if len(inter_channels) == len(self.config.decoder_stages):
            return inter_channels[idx]

        enabled_pos = enabled_indices.index(idx)
        return inter_channels[enabled_pos]

    def forward(self, x: Tensor) -> Tensor | dict[str, Tensor | list[Tensor]]:
        input_size = x.shape[-3:]
        x = self.stem(x)

        skips: list[Tensor] = []
        for idx, stage in enumerate(self.encoder):
            x = stage(x)
            skips.append(x)

            if idx < len(self.downsamplers):
                x = self.downsamplers[idx](x)

        x = self.bottleneck(x)

        aux_outputs: list[Tensor] = []
        decode_skips = list(reversed(skips[:-1]))

        for idx, (gate, upsample, stage, skip) in enumerate(
            zip(self.attention_gates, self.upsamplers, self.decoder, decode_skips)
        ):
            gated_skip = gate(skip, x)
            x = upsample(x, target_size=gated_skip.shape[-3:])
            x = self._merge_skip(x, gated_skip)
            x = stage(x)

            if self.config.deep_supervision and idx < len(self.aux_heads):
                aux = self.aux_heads[idx](x)
                aux_outputs.append(self._resize_to_input(aux, input_size))

        out = self.head(x)
        out = self._resize_to_input(out, input_size)

        if self.config.deep_supervision:
            return {"out": out, "aux": aux_outputs}

        return out



from lucid._tensor import Tensor
from lucid.types import _ArrayOrScalar, _DeviceType, Float32, Numeric

__all__ = ["Parameter", "Buffer"]


class Parameter(Tensor):
    def __init__(
        self,
        data: Tensor | _ArrayOrScalar,
        dtype: type | Numeric | None = None,
        device: _DeviceType = "cpu",
    ) -> None:
        orig_dtype: Numeric | bool | None = None
        if isinstance(data, Tensor):
            if isinstance(data.dtype, Numeric):
                orig_dtype = data.dtype
            data = data.data
        if (
            dtype is None
            and isinstance(orig_dtype, Numeric)
            and orig_dtype.base_dtype is float
            and orig_dtype.bits == 64
        ):
            dtype = Float32
        super().__init__(
            data, requires_grad=True, keep_grad=True, dtype=dtype, device=device
        )


class Buffer(Tensor):
    def __init__(
        self,
        data: Tensor | _ArrayOrScalar,
        dtype: type | None = None,
        device: _DeviceType = "cpu",
    ) -> None:
        if isinstance(data, Tensor):
            data = data.data
        super().__init__(
            data, requires_grad=False, keep_grad=False, dtype=dtype, device=device
        )
